"""
nilutility/numeric.py
$Id: numeric.py 2686 2021-08-18 11:24:59Z pe $

This module collects numerical helper functions that apply to differnt numeric
datatypes (float, Decimal, int, HexInt...), and can not be added to one specific
_helper module (e.g. float_helper).
"""

import re
import math
import string
from nilutility.ndecimal import Decimal
from nilutility.datatypes import HexInt

def frexp10(value):
    """
    Finds the mantissa m and exponent e of a number x such that x = m 10^e
    Parameters:
        value   value to be analyzed
    Returns:
        (mant, expon)
    """
    expon = int(math.floor(math.log10(abs(value))))
    mant = value/math.pow(10, expon)
    return (mant, expon)

def digits(value):
    """
    Returns the minimum and maximum digit used in a float, decimal or HexInt
    value.
    Parameters:
        value   float, Decimal or HexInt value (None allowed)
    Returns:
        digits tulple (mindig, maxdig)
        returns (None, None) if a None value is passed
        returns (0, 0) in case of 0.0

        Be aware that the number of digits are numbered starting from
        1/-1 left/right of the comma! This means that 0 is is not a valid digit.
        This also implies that the digit number (before comma) is not equal to
        the exponent of 10 of the respective digit.

        Digit 0 is only used if the number needs no digits at all (only 0.0)
        Digit None is used for None values.
    Raises:
        ValueError if the string representation does not fit to
                   float/Decimal behaviour.
        TypeError  if the type is not float, Decimal or HexInt
                   The TypeError will have an additional new_type arttribute
    """
    maxdig = None
    mindig = None
    if value is None:
        return (None, None)
    elif isinstance(value, HexInt):
        return _digits_hexint(value)
    elif isinstance(value, float):
        # special case, float is printed in full precision using repr
        # str shortens the output string in some cases.
        # However, repr cannot be used for other types - Decimal in
        # this case
        valstr = repr(value)
    elif isinstance(value, (Decimal, int)):
        valstr = str(value)
    else:
        err = TypeError('unsupported type')
        err.new_type = type(value)
        raise err
    reg = re.match(\
        r"^-?([0-9]*)(\.([0-9]*?[1-9]?)0*)?([Ee]([\+-][0-9]*))?$",
        valstr)
    if not reg:
        raise ValueError('illegal value: ' + valstr)
    vk_ = len(reg.group(1)) if reg.group(1) != None and \
                               reg.group(1) != '0' else 0
    nk_ = len(reg.group(3)) if reg.group(3) != None and \
                               reg.group(3) != '0' else 0
    if vk_ == 0 and nk_ == 0:
        # this meand value is 0.0
        return (0, 0)
    exp = int(reg.group(5)) if reg.group(5) != None else 0
    if nk_ == 0:
        # special case: if no digits after comma and some 0 before comma:
        reg2 = re.search('(0*)$', reg.group(1))
        mindig = exp + len(reg2.group(1)) + 1
        # reg.group(1)==None cannot happen (only if val was 0.0; caught before)
        if mindig <= 0:
            # correct mindig if calculated w/o nk_ but with a negative exponent
            mindig -= 1
    else:
        mindig = exp - nk_
        if nk_ and mindig >= 0:
            # zero is no digit number. If crossing zero, we nedd to add one
            mindig += 1
    if vk_ == 0:
        # special case: if no digits before comma
        reg2 = re.search('^(0*)', reg.group(3))
        maxdig = exp - len(reg2.group(1)) - 1
        # reg.group(3)==None cannot happen (only if val was 0.0; caught before)
    else:
        maxdig = exp + vk_
        if vk_ and maxdig <= 0:
            # zero is no digit number. If crossing zero, we nedd to subtract one
            maxdig -= 1
    return (mindig, maxdig)

def digits_stat(valuelist):
    """
    Calculates statistics about the values in a list of numeric values.
    digist_stat is designed and tested for float or Decimal values.
    Calculates minimum value, maximum value, minimum digit, maximum digit,
    maximum precision used.
    Parameters:
        valuelist  list of float or Decimal values (None allowed)
    Returns:
        statistics tulple (minval, maxval, mindig, maxdig, maxprec)
        (None, None, None, None, 0) if ONLY None values occurred (no data)
        (0.0, 0.0, 0, 0, 0) if 0.0 were the ONLY valid values
        (HexInt(0), HexInt(0), 0, 0, 0) if 0x0 were the ONLY valid values
    Raises:
        ValueError if the string representation does not fit to
                   float/Decimal behaviour.
        TypeError  if the type is not float or Decimal or HexInt
    """
    valuelist = [d for d in valuelist if d is not None]
    if not valuelist:
        return (None, None, None, None, 0)
    if isinstance(valuelist[0], HexInt):
        return _digits_stat_hexint(valuelist)
    t_maxdig = None
    t_mindig = None
    t_maxprec = 0
    t_minval = None
    t_maxval = None
    for val in valuelist:
        (mindig, maxdig) = digits(val)
        prec = maxdig - mindig
        if (mindig < 0 and maxdig < 0) or \
           (mindig > 0 and maxdig > 0):
            # e.g. 123., 1012   => mindig=1, maxdig=4, 4-1 = 3 (wrong)
            # but  123., 1012.1 => mindig=-1, maxdig=4, 4-(-1) = 5 (correct),
            # (it's all because we do NOT use a digit number 0)
            prec += 1
        t_maxprec = max(t_maxprec, prec)
        if t_mindig is None or t_mindig == 0:
            t_mindig = mindig
        elif mindig != 0:
            t_mindig = min(t_mindig, mindig)
        if t_maxdig is None or t_maxdig == 0:
            t_maxdig = maxdig
        elif maxdig != 0:
            t_maxdig = max(t_maxdig, maxdig)
        t_minval = val if t_minval is None else min(t_minval, val)
        t_maxval = val if t_maxval is None else max(t_maxval, val)
    return (t_minval, t_maxval, t_mindig, t_maxdig, t_maxprec)

def int2base(i, base, alphabet=string.digits+string.ascii_letters):
    """
    Convert an integer 'i' to a string representation in the number system of
    the base 'base'.
    Parameters:
        i         integer to be converted
        base      target number system
        alphabet  digit characters to be used
    Returns
        number as string the base system
    """
    if base > len(alphabet):
        raise ValueError(
            'not enough alphabet chars defined for number system {}'.format(
                base))
    if i < 0:
        sign = -1
    elif i == 0:
        return alphabet[0]
    else:
        sign = 1

    i *= sign
    digits = []

    while i:
        digits.append(alphabet[int(i % base)])
        i = int(i / base)

    if sign < 0:
        digits.append('-')

    digits.reverse()

    return ''.join(digits)

def _digits_hexint(value):
    """
    Returns the minimum and maximum digit used in a HexInt value.
    digits is designed and tested for float or Decimal values.
    Parameters:
        value   HexInt value (None is not allowed)
    Returns:
        digits tulple (mindig, maxdig)
        returns (0, 0) in case of 0x0

        Be aware that the number of digits are numbered starting from
        1/-1 left/right of the comma! This means that 0 is is not a valid digit.
        This also implies that the digit number (before comma) is not equal to
        the exponent of 16 of the respective digit.

        Digit 0 is only used if the number needs no digits at all (only 0.0)
    Raises:
        TypeError  if the type is not HexInt
                   The TypeError will have an additional new_type arttribute
    """
    if not isinstance(value, HexInt):
        err = TypeError('unsupported type')
        err.new_type = type(value)
        raise err
    if value == 0:
        return(0, 0)
    valstr = str(value)[2:]  # string representation without leading '0x'
    maxdig = len(valstr)
    mindig = len(valstr) - len(valstr.rstrip('0')) + 1
    return (mindig, maxdig)

def _digits_stat_hexint(valuelist):
    """
    For HexInt variables:
    Calculates statistics about the numerics of on variable (minimum digit,
    maximum digit, maximum precision used).
    Parameters:
        valuelist   list of values (all HexInt), no None values
    Returns:
        statistics tulple (minval, maxval, mindig, maxdig, maxprec)
        (HexInt(0), HexInt(0), 0, 0, 0) if 0x0 were the ONLY valid values
    Raises:
        TypeError if a non HexInt value is found in the variable.
    """
    t_maxdig = None
    t_mindig = None
    t_maxprec = 0
    t_minval = None
    t_maxval = None
    for val in valuelist:
        if not isinstance(val, HexInt):
            raise TypeError('non hex value in HexInt valuelist')
        # mindig is always 1 if there is at least one valid value != 0
        # if only 0x0, mindig is 0
        (mindig, maxdig) = _digits_hexint(val)
        t_mindig = max(t_mindig, mindig) if t_mindig is not None else mindig
        t_maxdig = max(t_maxdig, maxdig) if t_maxdig is not None else maxdig
        prec = 0 if val == 0 else maxdig-mindig+1
        t_maxprec = max(t_maxprec, prec)
        t_minval = min(t_minval, val) if t_minval is not None else val
        t_maxval = max(t_maxval, val) if t_maxval is not None else val
    return (t_minval, t_maxval, t_mindig, t_maxdig, t_maxprec)



"""
ebas/nasa_ames/read/parse_ebasmetadata.py
$Id: parse_ebasmetadata.py 2035 2018-05-22 14:13:03Z pe $

parser for file input functionality for EBAS NASA Ames module
parses ebas tag=value pairs from NasaAmes1001 object and builds NasaAmes

The basic parsing (and very fundamental value checks) are done by this module.
What is not done here:
 - More advanced value checks that need validation data from the DB
   --> NasaAmesToDomain
 - Value checks which depend on other metadata elements can not be done, those
   need to be checked after reading the complete header; see:
   --> NasaAmesBase.check_interdependent_metadata

History:
V.1.0.0  2013-06-22  pe  initial version

"""

import re
import datetime
from .parse_base import NasaAmesPartialReadParserBase
# pylint: disable=F0401
# F0401: Unable to import 'ebasmetadata' (pylint problem)
from ....ebasmetadata import EbasMetadataError, EbasMetadataWarning, \
    DatasetCharacteristicList, DCError, DCWarning
from nilutility.datatypes import DataObject
from ...basefile.base import  EbasMetadataInvalid


class NasaAmesPartialReadParserEbasMetadata(# pylint: disable=R0901, W0223,
        NasaAmesPartialReadParserBase):
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Parser for Nasa Ames I/O object.
    This is a base class for NasaAmesPartialReadParserVariables and
    NasaAmesPartialReadParser
    """

    def _parse_ebasmetadata_vname_positional(self, comp_name, unit, lnum,
                                             var_index, ignore_parameter=False):
        """
        Parse positional metadata elements from VNAME: comp_name, unit.
        Parameters:
            comp_name, unit   component name and unit from vname line
            lnum              line number (for error reporting)
            var_index         variable number in file
            ignore_parameter
                    ignore errors related to paramters and units this is needed
                    when a file with non standard vnames should be processed
                    without importing it into the ebas.domain.
        Returns:
            (comp_name, unit), invalid values are changed to EbasMetadataInvalid
        """
        # check component
        try:
            self.internal.ebasmetadata.parse_component(comp_name)
        except EbasMetadataError as excpt:
            msg = "Variable {}: {}".format(var_index, str(excpt))
            if ignore_parameter:
                self.warning(msg, lnum=lnum)
            else:
                self.error(msg, lnum=lnum)
                # set to EbasMetadataInvalid in order to avoid errors that
                # result from this one here.
                # E.g. mandatory missing, or things checked in
                # check_interdependent
                comp_name = EbasMetadataInvalid(comp_name)
        except EbasMetadataWarning as excpt:
            comp_name = excpt.retval
            msg = "Variable {}: {}".format(var_index, str(excpt))
            self.warning(msg, lnum=lnum)

        # check unit
        try:
            self.internal.ebasmetadata.parse_unit(unit)
        except EbasMetadataError as excpt:
            msg = "Variable {}: {}".format(var_index, str(excpt))
            if ignore_parameter:
                self.warning(msg, lnum=lnum)
            else:
                self.error(msg, lnum=lnum)
                # set to EbasMetadataInvalid in order to avoid errors that
                # result from this one here.
                # E.g. mandatory missing, or things checked in
                # check_interdependent
                unit = EbasMetadataInvalid(unit)

        return (comp_name, unit)

    def _parse_ebasmetadata(self, tag, value, lnum, var_index=None,
                            ignore_parameter=False):
        """
        Parse a tag=value pair from EBAS specific metadata.
        This can be either ebas core metadata or dataset characteristics.
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        metadata_tags = self.internal.ebasmetadata.metadata_tags
        # special case for QA metadata:
        if self._parse_ebasmetadata_qa(tag, value, lnum, var_index):
            pass
        elif tag in metadata_tags:
            # this is a core ebas metadata element (not a characteristic)
            if var_index != None and not metadata_tags[tag]['vname']:
                # [vname] == 0: not valid in VNAME
                self.error(tag + ": is not legal in VNAME", lnum=lnum)
                return # stop parsing this element
            elif var_index is None and not metadata_tags[tag]['main']:
                # [main] == 0: not valid for NCOML
                self.error(
                    tag + ": is not legal in main metadata block [NCOML]",
                    lnum=lnum)
                return # stop parsing this element
            if 'renamed_tag' in metadata_tags[tag] and \
               metadata_tags[tag]['renamed_tag']:
                self.warning(
                    "Legacy metadata element '{}' translated to '{}'"\
                    .format(tag, metadata_tags[tag]['renamed_tag']),
                    lnum=lnum)
            if self.metadata.datadef == 'EBAS_1' and value == 'NA':
                value = ''
            # Special Cases:
            parse_ebasmetadata_specific = {
                'Originator': self._parse_ebasmetadata_person,
                'Submitter': self._parse_ebasmetadata_person,
                'Timeref': self._parse_ebasmetadata_timeref,
                'Input dataset': self._parse_ebasmetadata_multiple_list,
                'Software': self._parse_ebasmetadata_multiple_list,
            }
            if tag in parse_ebasmetadata_specific:
                parse_ebasmetadata_specific[tag](tag, value, lnum, var_index)
            else:
                store_value = self._parse_ebasmetadata_generic(tag, value, lnum,
                                                               var_index)
                emptyval = None
                self._set_metadata(tag, store_value, lnum, var_index,
                                   emptyval=emptyval)
        else:
            # else, we assume it's a characteristic (validity can't be checked
            # without master data from database)
            self._parse_characteristics(tag, value, lnum, var_index,
                                        ignore_parameter=ignore_parameter)

    def _parse_ebasmetadata_qa(self, tag, value, lnum, var_index):
        """
        Parse tag=value pair from EBAS specific core metadata (one
        that can be processed generically by the ebasmetadata module).
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            True      was QA meatdata, parsed OK
            False     not QA metadata, caller needs to keep on processing
        """
        # get generic tag and QA number
        reg = re.match('^QA([0-9]*)( .*)$', tag)
        if reg:
            # this is a QA metadata element:
            generic_tag = "QA" + reg.group(2)
            qa_number = int(reg.group(1)) if reg.group(1) else -1
        else:
            return False  # not QA metadata

        metadata_tags = self.internal.ebasmetadata.metadata_tags
        if not generic_tag in metadata_tags:
            return False

        store_value = self._parse_ebasmetadata_generic(generic_tag, value, lnum,
                                                       var_index)

        key = self.internal.ebasmetadata.metadata_tags[generic_tag]['key']
        qa_ = self.get_qa_by_number(qa_number, var_index)
        if qa_ is not None and key in qa_ and qa_[key] is not None:
            self.error(
                "multiple definitions for '{}'".format(tag), lnum=lnum)
            return
        # set:
        self.set_qa(key, store_value, qa_number, var_index)

        return True

    def _parse_ebasmetadata_generic(self, tag, value, lnum, var_index):
        # pylint: disable-msg=R0912
        # R0912: Too many branches
        """
        Parse tag=value pair from EBAS specific core metadata (one
        that can be processed generically by the ebasmetadata module).
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            parsed (and converted, manipulated) value
        """
        occ = 'main' if var_index is None else 'vname'
        elem = self.internal.ebasmetadata.metadata_tags[tag]
        # parse value
        if value == '' or value is None:
            return None
        # check if some special params are set for default parsing
        params = {key+'_': elem[key] for key in ('unit', 'type', 'range', 'len')
                  if key in elem and elem[key] is not None}
        try:
            if 'parser' in elem and elem['parser']:
                value = elem['parser'](value)
            elif params:
                value = self.internal.ebasmetadata.parse_default(
                    value, **params)
            # else, just take the string as is
        except EbasMetadataError as excpt:
            self.error(tag + ": " + str(excpt), lnum=lnum)
            if elem[occ] & 8:
                self.internal.read_tmp.hea_error = True
            # set to EbasMetadataInvalid in order to avoid errors that
            # result from this one here.
            # E.g. mandatory missing, or things checked in
            # check_interdependent
            return EbasMetadataInvalid(value)
        except EbasMetadataWarning as excpt:
            value = excpt.retval
            self.warning(tag + ": " + str(excpt), lnum=lnum)
        return value

    def _parse_ebasmetadata_multiple_list(self, tag, value, lnum, var_index):
        # pylint: disable-msg=R0912
        # R0912: Too many branches
        """
        Parse tag=value pair for metadata which support multiple elements, but
        but only in global metadata.
        Currently only 'Input dataset'.
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
                          Not supported
        Returns:
            None
        """
        if value == '' or value is None:
            return
        if var_index is None:
            elem = self.internal.ebasmetadata.metadata_tags[tag]
            # check if some special params are set for default parsing
            params = {key+'_': elem[key]
                      for key in ('unit', 'type', 'range', 'len')
                      if key in elem and elem[key] is not None}
            try:
                if 'parser' in elem and elem['parser']:
                    value = elem['parser'](value)
                elif params:
                    value = self.internal.ebasmetadata.parse_default(
                        value, **params)
                # else, just take the string as is
            except EbasMetadataError as excpt:
                self.error(tag + ": " + str(excpt), lnum=lnum)
                if elem['main'] & 8:
                    self.internal.read_tmp.hea_error = True
            except EbasMetadataWarning as excpt:
                value = excpt.retval
                self.warning(tag + ": " + str(excpt), lnum=lnum)
            if not self.metadata[elem['key']]:
                self.metadata[elem['key']] = []
            self.metadata[elem['key']].append(value)
        else:
            # not supported, should not be set in ebasmetadata
            raise RuntimeError()

    def _set_metadata(self, tag, value, lnum, var_index, emptyval=None):
        """
        Set a metadata value for a generic ebas metadata element.
        Check if element has been set before (log error if set already)
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        key = self.internal.ebasmetadata.metadata_tags[tag]['key']
        if var_index is None:
            occurrence = 'main'
            metadata = self.metadata
        else:
            occurrence = 'vname'
            metadata = self.variables[var_index].metadata
        # check if multiply set:
        if key in metadata and \
             (metadata[key] != emptyval or occurrence == 'vname') and \
             (key != 'revdate' or
              (occurrence == 'main' and self.internal.read_tmp.revdate_read)):
            self.error(
                "multiple definitions for '{}'".format(tag), lnum=lnum)
        elif occurrence == 'main' and key == 'revdate':
            self.internal.read_tmp.revdate_read = True
            if isinstance(value, datetime.datetime) and \
               isinstance(self.metadata['revdate'], datetime.datetime) and \
               value.date() != self.metadata['revdate'].date():
                self.error(
                    "'{}' ({}...) must be the same day as RDATE specified in "
                    "line 7 ({})".format(
                        tag, datetime.datetime.strftime(value, "%Y%m%d"),
                        datetime.datetime.strftime(
                            self.metadata['revdate'], "%Y %m %d")),
                    lnum=lnum)
            if isinstance(value, datetime.datetime):
                metadata[key] = value
        elif key:
            metadata[key] = value

    def _parse_ebasmetadata_person(self, tag, value, lnum, var_index):
        """
        Parse EBAS person metadata (Submitter or Originator).
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        if value == '' or value is None:
            return None
        metadata_tags = self.internal.ebasmetadata.metadata_tags
        key = metadata_tags[tag]['key']
        if var_index is None:
            try:
                psd = metadata_tags[tag]['parser'](value)
            except EbasMetadataError as excpt:
                self.error(tag + ": " + str(excpt), lnum=lnum)
                return  # stop parsing this line
            except EbasMetadataWarning as excpt:
                psd = excpt.retval
                self.warning(tag + ": " + str(excpt), lnum=lnum)
            for pers in self.metadata[key]:
                if pers.PS_LAST_NAME == psd['PS_LAST_NAME'] and \
                   pers.PS_FIRST_NAME == psd['PS_FIRST_NAME']:
                    pers.update(psd)
                    self.logger.debug(u'{}={}'.format(tag, str(pers)))
                    break
            else:
                self.error(tag + u": '{}, {}'".\
                    format(psd['PS_LAST_NAME'], psd['PS_FIRST_NAME']) +\
                    u" not found in {} (line {})".\
                    format('ONAME' if tag == 'Originator' else 'SNAME',
                           2 if tag == 'Originator' else 4), lnum=lnum)
        else:
            # do we want to support Submitter / Originator in VNAME?
            raise NotImplementedError()

    def _parse_ebasmetadata_timeref(self, tag, value, lnum, var_index):
        """
        Parse an exceptional tag=value pair from EBAS specific metadata (one
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOM block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        if value == '' or value is None:
            return None
        metadata_tags = self.internal.ebasmetadata.metadata_tags
        if var_index != None:
            # this should have been caught by metadata_tags[tag]['vname']
            raise ValueError()
        try:
            (timeref, timeoffset) = metadata_tags[tag]['parser'](value)
        except EbasMetadataError as excpt:
            self.error(tag + ": " + str(excpt), lnum=lnum)
            occ = 'main' if var_index is None else 'vname'
            if metadata_tags[tag][occ] & 8:
                self.internal.read_tmp.hea_error = True
            if metadata_tags[tag][occ] & 4 or metadata_tags[tag][occ] & 8:
                # set to empty, in order to avoid 'mandatory missing' errors
                self._set_metadata(tag, '', lnum, var_index)
            return
        except EbasMetadataWarning as excpt:
            (timeref, timeoffset) = excpt.retval
            self.warning(tag + ": " + str(excpt), lnum=lnum)
        self._set_metadata(tag, timeref, lnum, var_index)
        self.internal.timeoffset = timeoffset

    def _parse_characteristics(self, tag, value, lnum, var_index=None,
                               ignore_parameter=False):
        """
        Parses and sets characteristics.
        Parameters:
            tag, value    tag=value pair from VNAME line or NCOML block
            lnum          line number (for error reporting)
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOML block)
        Returns:
            None
        """
        ft_type = ''
        co_comp_name = ''
        # get fi_type and co_comp_name from main metadata:
        if 'instr_type' in self.metadata:
            ft_type = self.metadata.instr_type
        if 'comp_name' in self.metadata:
            co_comp_name = self.metadata.comp_name
        # get the metadata set we should use for adding DC
        if var_index != None:
            metadata = self.variables[var_index].metadata
            # set variable_specific ft_type and co_comp_name
            if 'instr_type' in metadata:
                ft_type = metadata.instr_type
            if 'comp_name' in metadata:
                co_comp_name = metadata.comp_name
        else:
            metadata = self.metadata
            
        if isinstance(ft_type, EbasMetadataInvalid):
            ft_type = None  # set to none if EbasMetadataInvalid
        if isinstance(co_comp_name, EbasMetadataInvalid):
            co_comp_name = None  # set to none if EbasMetadataInvalid

        # init characteristics list if needed:
        if 'characteristics' not in metadata or \
           metadata.characteristics is None:
            metadata.characteristics = DatasetCharacteristicList()

        # create DC object (parse and validate)
        try:
            metadata.characteristics.add_parse(
                tag, value, ft_type, co_comp_name,
                data_level=self.metadata.datalevel)
        except DCError as excpt:
            if ignore_parameter:
                self.warning(str(excpt), lnum=lnum)
            else:
                self.error(str(excpt), lnum=lnum)
            return
        except DCWarning as excpt:
            self.warning(str(excpt), lnum=lnum)
            return



"""
chemical elements properties
"""

# Fundamental constants for chemical elements
# Constants are:
# AtomicWeight: the ratio of the mass of one atom to 1/12 of the mass of one
#               carbon-12 atom (unit: dimensionless)
# MolarMass: the mass of one mole of the element
#
# Annotation: the numbers for AtomicWeight and MolarMass are by definition
# equal. We use different constants anyway because they have different
# dimension.
ELEMENTS = \
{
    "C":
    {
        "MolarMass": 12.011,  # molar mass (g/mol)
        "AtomicWeight": 12.011,  # (dimensionless)
        # Source: Conventional Atomic Weights 2013. Commission on Isotopic
        # Abundances and Atomic Weights: [12.0096, 12.0116]
    },
    "Ca":
    {
        "MolarMass": 40.078,  # molar mass (g/mol)
        "AtomicWeight": 40.078, # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515%2Fpac-2015-0305)
    },
    "Cl":
    {
        "MolarMass": 35.45,  # molar mass (g/mol)
        "AtomicWeight": 35.45, # (dimensionless)
        # Source: Conventional Atomic Weights 2013.
        # (isotope mix is highly variable due to regional source of the material
        # the conventional atomic weight specifies a global average
        # https://doi.org/10.1515%2Fpac-2015-0305)
        # Atomic weights of the elements 2013 (IUPAC Technical Report)
        # [35.446, 35.457], conventional: 35.45
    },
    "F":
    {
        "MolarMass": 18.998403163,  # molar mass (g/mol)
        "AtomicWeight": 18.998403163,  # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515/pac-2015-0305
    },
    "H":
    {
        "MolarMass": 1.008,  # molar mass (g/mol)
        "AtomicWeight": 1.008,  # (dimensionless)
        # Source: Conventional Atomic Weights 2013. Commission on Isotopic
        # Abundances and Atomic Weights: [1.00784-1.00811]
    },
    "K":
    {
        "MolarMass": 39.0983,  # molar mass (g/mol)
        "AtomicWeight": 39.0983, # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515%2Fpac-2015-0305)
    },
    "Mg":
    {
        "MolarMass": 24.305,  # molar mass (g/mol)
        "AtomicWeight": 24.305, # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515%2Fpac-2015-0305)
        # Conventional atomic weight
    },
    "N":
    {
        "MolarMass": 14.007,  # molar mass (g/mol)
        "AtomicWeight": 14.007,  # (dimensionless)
        # Source: Conventional Atomic Weights 2015. Commission on Isotopic
        # Abundances and Atomic Weights: [14.00643, 14.00728]
    },
    "Na":
    {
        "MolarMass": 22.98976928,  # molar mass (g/mol)
        "AtomicWeight": 22.98976928, # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515%2Fpac-2015-0305)
    },
    "O":
    {
        "MolarMass": 15.9994,  # molar mass (g/mol)
        "AtomicWeight": 15.9994,  # (dimensionless)
        # Source: Conventional Atomic Weights 2015. Commission on Isotopic
        # Abundances and Atomic Weights: [15.99903, 15.99977]
    },
    "P":
    {
        "MolarMass": 30.973761998,  # molar mass (g/mol)
        "AtomicWeight": 30.973761998,  # (dimensionless)
        # Source: Atomic weights of the elements 2013 (IUPAC Technical Report)
        # https://doi.org/10.1515/pac-2015-0305
    },
    "S":
    {
        "MolarMass": 32.06,  # molar mass (g/mol)
        "AtomicWeight": 32.06, # (dimensionless)
        # Source: Conventional Atomic Weights 2013.
        # (isotope mix is highly variable due to regional source of the material
        # the conventional atomic weight specifies a global average
        # https://doi.org/10.1515/pac-2015-0305)
        # Standard Atomic Weights 2015. Commission on Isotopic
        # Abundances and Atomic Weights: [32.059-32.076]
    },
}




#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='ebas-io',
      version='4.04.03',
      description='EBAS file I/O',
      author='Paul Eckhardt, NILU',
      author_email='pe@nilu.no',
      license='agpl 3.0',
      #url='https://www.python.org/sigs/distutils-sig/',
      packages=find_packages(exclude=["*.test", "*.test.*"]),                    

      # add pkl files for offline masterdata:
      include_package_data=True,
      package_data={'ebas.domain.masterdata.offline_masterdata': ['*.pkl']},
      install_requires=['python-dateutil', 'datadiff'],
     )




"""
$Id: sm.py 2466 2020-07-10 13:20:48Z pe $

EBAS Masterdata Class for standard methods

This module implements the class EbasMasterSM.
"""

from .offline_masterdata import SMOfflineMasterData
from .base import EbasMasterBase

class EbasMasterSM(EbasMasterBase, SMOfflineMasterData):
    """
    Domain Class for standard method masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling standard methods and checking them against master data.
    Standard method master data are retrieved from database or from offline
    storage when no database access is possible.
    """

    # read offline masterdata for SM (standard methods)
    # Those are fallback values, will be read from database as soon as possible.
    SMOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read sm masterdata if dbh is provided.
        """
        SMOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    @classmethod
    def exist_standardmethod(cls, standard_method):
        """
        Check if the standard_method exists (regardless of FT).
        Parameters:
            standard_method    standard method code to be checked
        Returns:
            True/False
        """
        if any([key[1] == standard_method for key in list(cls.META.keys())]):
            return True
        return False



"""
atmos_phys unit conversion routines for electrical quantities.
"""

from nilutility.datatypes import DataObject
from .base import ConvertBaseFactor


class ConvertElectricCurrent(ConvertBaseFactor):
    """
    Conversion class for electric current.
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        factors = {
            ('mA', 'A'): 0.001,
            ('A', 'mA'): 1000.0,
        }

        if self.from_unit not in \
               ('A', 'mA') or \
           self.to_unit not in \
               ('A', 'mA') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(
            factor=factors[(self.from_unit, self.to_unit)])

class ConvertConductivity(ConvertBaseFactor):
    """
    Conversion class for electric conductivity.
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        factors = {
            ('mS/m', 'uS/cm'): 10.0,
            ('uS/cm', 'mS/m'): 0.1,
        }

        if self.from_unit not in \
               ('mS/m', 'uS/cm') or \
           self.to_unit not in \
               ('mS/m', 'uS/cm') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(
            factor=factors[(self.from_unit, self.to_unit)])




"""
NOAA NMHC Flask data file format
   partial class for reading

$Id: read.py 2706 2021-09-14 08:57:09Z pe $
"""

from .base import NOAA_NMHC_Flask_Base, NOAA_NMHC_Flask_Rawdata_Base, \
    NOAA_NMHC_Flask_Header_Base, NOAA_NMHC_Flask_Data_Base, \
    NOAA_NMHC_Flask_HeaderContact_Base, NOAA_NMHC_Flask_HeaderDescription_Base,\
    NOAA_NMHC_Flask_Error
from six import string_types
import re
import datetime
from nilutility.datetime_helper import datetime_parse_iso8601

class NOAA_NMHC_Flask_ReadError(NOAA_NMHC_Flask_Error):
    """
    Exception class for read errors.
    """
    pass


class NOAA_NMHC_Flask_Rawdata_Read(NOAA_NMHC_Flask_Rawdata_Base):
    """
    Partial class for reading raw data from the file.
    """
    def parse(self, raw_data):
        """
        Parses the raw data (parse number of header lines, split into header
        and data.
        """
        self.raw_data = raw_data
        reg = re.match(r'^# number_of_header_lines: *(\d*)$', raw_data[0])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError("not a NOAA_NMHC_Flask file")
        self.num_hea_lines = int(reg.group(1))
        if self.num_hea_lines > len(self.raw_data):
            raise NOAA_NMHC_Flask_ReadError(
                "less lines in file then specified number_of_header_lines in "
                "first line")
        self.header = self.raw_data[0:self.num_hea_lines]
        self.data = self.raw_data[self.num_hea_lines:]

class NOAA_NMHC_Flask_HeaderContact_Read(NOAA_NMHC_Flask_HeaderContact_Base):
    """
    Partial class for parsing the contact information from the file header.
    """
    def parse(self, index, rawhea):
        """
        Parses a contact information block from the file header.
        Parameters:
            index      start index of the contact block
            rawhea     raw header lines (header attribute from
                       NOAA_NMHC_Flask_Rawdata object)
        Returns:
            index of the line following the block
        Syntax of a contact block:
            # contact_parameter: IC4H10 ARL (Individual Flasks)\n
            # contact_name: Detlev Helmig\n
            # contact_telephone: (303) 492-2509\n
            # contact_email: Detlev.Helmig@colorado.edu\n
        """
        reg = re.match(r'^# contact_parameter: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# contact_parameter'".format(index + 1))
        self.parameter = reg.group(1)
        index += 1

        reg = re.match(r'^# contact_name: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# contact_name'".format(index + 1))
        self.name = reg.group(1)
        index += 1

        reg = re.match(r'^# contact_telephone: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# contact_telephone'".format(index + 1))
        self.telephone = reg.group(1)
        index += 1

        reg = re.match(r'^# contact_email: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# contact_email'".format(index + 1))
        self.email = reg.group(1)
        index += 1

        return index


class NOAA_NMHC_Flask_HeaderDescription_Read(
        NOAA_NMHC_Flask_HeaderDescription_Base):
    """
    Partial class for parsing the description information from the file header.
    """
    def parse(self, index, rawhea):
        """
        Parses a description information block from the file header.
        Parameters:
            index      start index of the contact block
            rawhea     raw header lines (header attribute from
                       NOAA_NMHC_Flask_Rawdata object)
        Returns:
            index of the line following the block
        Syntax of a description block:
            # description_site-code: zep\n
            # description_project-abbr: ccg_surface\n
            # description_strategy-abbr: flask\n
            # description_sample-constraints: date:1900-01-01,2016-12-31\n
            # description_creation-time: 2017-02-01 16:54:49\n
        """
        reg = re.match(r'^# description_site-code: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# description_site-code'".format(index + 1))
        self.site_code = reg.group(1)
        if not re.match(r'^[a-z]{3}$', self.site_code):
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: site_code must be 3 characters, lower case'".format(
                    index + 1))
        index += 1

        reg = re.match(r'^# description_project-abbr: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# description_project-abbr'".format(
                    index + 1))
        self.project_abbr = reg.group(1)
        index += 1

        reg = re.match(r'^# description_strategy-abbr: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# description_strategy-abbr'".format(
                    index + 1))
        self.strategy_abbr = reg.group(1)
        if (self.strategy_abbr != 'flask'):
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: only flask is supported for strategy_abbr".format(
                    index + 1))
            # other file content might need changes in the implementation?
        index += 1

        reg = re.match(r'^# description_sample-constraints: *(.*)$',
                       rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# description_sample-constraints'".format(
                    index + 1))
        self.sample_constraints = reg.group(1)
        index += 1

        reg = re.match(r'^# description_creation-time: *(.*)$', rawhea[index])
        if not reg:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: expected '# description_creation-time'".format(
                    index + 1))
        try:
            self.creation_time = datetime_parse_iso8601(
                reg.group(1).replace(" ","T"))
        except ValueError:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: illegal time string {}'".format(
                    index + 1, reg.group(1)))
        index += 1

        return index


class NOAA_NMHC_Flask_Header_Read(NOAA_NMHC_Flask_Header_Base):
    """
    Partial class for parsing the header part of the file.
    """
    def parse(self, rawhea):
        """
        Parses the header information from the raw data object
        (NOAA_NMHC_Flask_Rawdata).
        Parameters:
            rawhea     raw header lines (header attribute from
                       NOAA_NMHC_Flask_Rawdata object)
        Returns:
            None
        """
        from .__init__ import NOAA_NMHC_Flask_HeaderContact, \
            NOAA_NMHC_Flask_HeaderDescription, NOAA_NMHC_Flask_Data

        i = 1  # skip first line (number_of_header_lines)
        while  i < len(rawhea):
            if rawhea[i].startswith('# comment:'):
                start = i
                i += 1
                while i < len(rawhea):
                    if not rawhea[i].startswith('# comment:'):
                        break
                    i += 1
                block = rawhea[start:i]
                self.comments.append(block)
            elif rawhea[i].startswith('# contact_parameter: '):
                contact = NOAA_NMHC_Flask_HeaderContact()
                i = contact.parse(i, rawhea)
                self.contact.append(contact)
            elif rawhea[i].startswith('# collaborator_name: '):
                reg = re.match(r'^# collaborator_name: *(.*)$', rawhea[i])
                if not reg:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: collaborator_name could not be parsed".format(
                            i+1))
                self.collaborator_name.append(reg.group(1))
                i += 1
            elif rawhea[i].startswith('# collaborator_comment: '):
                reg = re.match(r'^# collaborator_comment: *(.*)$', rawhea[i])
                if not reg:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: collaborator_comment could not be parsed"\
                            .format(i+1))
                self.collaborator_comment.append(reg.group(1))
                i += 1
            elif rawhea[i].startswith('# collaborator_url: '):
                reg = re.match(r'^# collaborator_url: *(.*)$', rawhea[i])
                if not reg:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: collaborator_url could not be parsed".format(
                            i+1))
                self.collaborator_url.append(reg.group(1))
                i += 1
            elif rawhea[i].startswith('# collaborator_comment: '):
                reg = re.match(r'^# collaborator_comment: *(.*)$', rawhea[i])
                if not reg:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: collaborator_comment could not be parsed".format(
                            i+1))
                self.collaborator_comment.append(reg.group(1))
                i += 1
            elif rawhea[i].startswith('# description_site-code: '):
                if self.description:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: only one description block allowed".format(
                            i+1))
                description = NOAA_NMHC_Flask_HeaderDescription()
                i = description.parse(i, rawhea)
                self.description = description
            elif rawhea[i].startswith('# data_fields: '):
                if self.data_fields:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: only one line with data_fields allowed".
                        format(i+1))
                reg = re.match(r'^# data_fields: *(.*)$', rawhea[i])
                if not reg:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: error parsing data_fields".format(i+1))
                self.data_fields = reg.group(1).split()
                expected_data_fields = NOAA_NMHC_Flask_Data.DATA_FIELDS
                if self.data_fields != expected_data_fields:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}: expected data_fields {}, read {}".format(
                            i + 1, ', '.join(expected_data_fields),
                            ', '.join(self.data_fields)))
                i += 1
            else:
                self._logger.error(
                    "cannot parse header line '{}'".format(rawhea[i]))
                i += 1


class NOAA_NMHC_Flask_Data_Read(NOAA_NMHC_Flask_Data_Base):
    """
    Partial class for reading the data part of the file.
    """

    def parse(self, raw):
        """
        Parses the data section of the file.
        Parameters:
            raw    NOAA_NMHC_Flask_Rawdata object  
        Returns:
            None
        """
        for i in range(len(raw.data)):
            self._parse_data_line(len(raw.header)+i+1, raw.data[i])

    def _parse_data_line(self, lnum, data_line):
        """
        Parses one line of data and adds the data to the object.
        Parameters:
            lnum         line number (in file, 1 based, used for messages)
            data_line    raw string with one line of data
        Returns:
            None
        Raises:
            raise NOAA_NMHC_Flask_ReadError("not a NOAA_NMHC_Flask file")xx
        """
        data_list = data_line.split()
        if len(data_list) != 27:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: data line contains {} elements (27 expected)".format(
                    lnum, len(data_list)))
        for (i, elem) in enumerate(self.__class__.DATA_FIELDS):
            # do some type converstion:

            if elem in ('sample_year', 'sample_month', 'sample_day',
                        'sample_hour', 'sample_minute', 'sample_seconds',
                        'analysis_year', 'analysis_month', 'analysis_day',
                        'analysis_hour', 'analysis_minute', 'analysis_seconds'):
                # convert to integer
                try:
                    data_list[i] = int(data_list[i])
                except ValueError:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}, col {} ({}): expected integer value, "
                        "found '{}'".format(lnum, i, elem, data_list[i]))
            if elem in ('analysis_value', 'analysis_uncertainty',
                        'sample_latitude', 'sample_longitude',
                        'sample_altitude', 'sample_elevation',
                        'sample_intake_height'):
                # convert to float
                try:
                    data_list[i] = float(data_list[i])
                except ValueError:
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}, col {} ({}): expected float value, "
                        "found '{}'".format(lnum, i, elem, data_list[i]))
            # apply some additional actions or checks for some elements:
            if elem in ('analysis_value', 'analysis_uncertainty'):
                # set to None if missing -999.990 or -999.999
                if data_list[i] in (-999.990, -999.999):
                    data_list[i] = None
                elif data_list[i] < -5:
                    # Sanity checks: do not accept low negative values
                    # Maybe this limit needs to be adjusted
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}, col {} ({}): very low negative value "
                        "found '{}'".format(lnum, i, elem, data_list[i]))
                elif data_list[i] in (999.990, 999.999):
                    raise NOAA_NMHC_Flask_ReadError(
                        "line {}, col {} ({}): suspicious (missing?) value "
                        "found '{}'".format(lnum, i, elem, data_list[i]))
            # append to data attributes:
            self.__dict__[elem].append(data_list[i])

        # check elevation + intake_height == altitude:
        if abs(self.sample_altitude[-1] - self.sample_elevation[-1] - \
               self.sample_intake_height[-1]) > 0.01:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}: sample_elevation + sample_intake_height is not equal"
                "sample_altitude ({}+{}!={})".format(
                    lnum, self.sample_elevation[-1],
                    self.sample_intake_height[-1], self.sample_altitude[-1]))
        # set sample_time
        try:
            self.sample_time.append(datetime.datetime(
                self.sample_year[-1], self.sample_month[-1],
                self.sample_day[-1],
                self.sample_hour[-1], self.sample_minute[-1],
                self.sample_seconds[-1]))
        except ValueError as expt:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}, col 2-7 (sample time): {}".format(
                    lnum, expt.message))
        # set analysis_time
        try:
            self.analysis_time.append(datetime.datetime(
                self.analysis_year[-1], self.analysis_month[-1],
                self.analysis_day[-1],
                self.analysis_hour[-1], self.analysis_minute[-1],
                self.analysis_seconds[-1]))
        except ValueError as expt:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}, col 16-21 (analysis time): {}".format(
                    lnum, expt.message))

        # sample_time < analysis_time
        # 2021-05: initially, there were errors in many files.
        #          INSTAAR/NOAA investigated, conclusion was that analysis date
        #          is wrong, sample date OK. Changed to sample date in those
        #          cases.
        # --> accept analysis time == sample time for this reason
        if self.analysis_time[-1] < self.sample_time[-1]:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}, col 2-7, 16-21 (sample time and analysis time): "
                "sample time must be before analysis time".format(
                    lnum))

        if len(self.sample_time) > 1 and \
           self.sample_time[-2] > self.sample_time[-1]:
            raise NOAA_NMHC_Flask_ReadError(
                "line {}, col 2-7 (sample time): sample time must be "
                "monotonically increasing".format(lnum))


class NOAA_NMHC_Flask_Read(NOAA_NMHC_Flask_Base):
    """
    Partial class for reading a file.
    """

    def read(self, filespec):
        """
        Reads the file into the object.
        Parameters:
            filespec    file name (incl. path) or file like object (stream)
        Returns:
            None
        Raises:
            IOError                 file open error
            NOAA_NMHC_Flask_Error (or any derived class hereof)
                on any unrecoverable read error
        """
        if isinstance(filespec, string_types):
            self.filename = filespec
            with open(self.filename, "Ur") as self.file:
                self._logger.info("reading file " + self.filename)
                self._read()
        else:
            self.filename = None
            self._logger.info("reading from stream")
            self.file = filespec
            self._read()

    def _read(self):
        """
        Reads the file from opened file handle.
        This method is only needed to extract the reading functionality in order
        to wrap the 'with open(filename)' around it in the read() method.
        Parameters:
            None
        Returns:
            None
        Raises:
            NOAA_NMHC_Flask_Error (or any derived class hereof)
                on any unrecoverable read error
        """
        from .__init__ import NOAA_NMHC_Flask_Rawdata, NOAA_NMHC_Flask_Header, \
            NOAA_NMHC_Flask_Data
        self.raw = NOAA_NMHC_Flask_Rawdata()
        self.raw.parse(self.file.readlines())
        self.header = NOAA_NMHC_Flask_Header()
        self.header.parse(self.raw.header)
        self.data = NOAA_NMHC_Flask_Data(self.raw.num_hea_lines+1)
        self.data.parse(self.raw)
        self.check()
        #self.data.parse()



"""
$Id: sqlite.py 2599 2021-01-27 16:27:04Z pe $

Helper module fro accessing sqlite3 databases.
"""

import logging
import uuid
import six
from textwrap import dedent
import sqlite3
from sqlite3 import OperationalError  # @UnusedImport pylint: disable=W0611
# W0611: Unused OperationalError imported from sqlite3 (unused-import)
# --> needed for re-export
from nilutility.datatypes import DataObject
from .base import DbBase

class SqliteDb(DbBase):
    """
    Base class for sqlite3 database access.
    """

    # Used by base class
    DB_MAXLEN_IDENTIFIER = None

    def __init__(self, filename):
        """
        Initialize DB connection. Open DB.
        Parameters:
            filename   filename for the database.
        Returns:
            None
        Raises:
            e.g. sqlite3.OperationalError, most usual exceptions to catch by
            caller:
                except OperationalError as excpt:
                    if excpt.message == 'unable to open database file':
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.filename = filename
        self.dbh = None
        self.tmp_addsql_crit_tables = []
        self.open()

    def open(self):
        """
        Open the database. If the file does not exists, create anew one.
        Parameters:
            None
        Returns:
            None
        Raises:
        Raises:
            e.g. sqlite3.OperationalError, most usual exceptions to catch by
            caller:
                except OperationalError as excpt:
                    if excpt.message == 'unable to open database file':
        """
        self.logger.info('open database %s', self.filename)
        self.dbh = sqlite3.connect(
            self.filename,
            detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)
        sqlite3.register_adapter(bool, int)
        sqlite3.register_converter("BOOLEAN", lambda v: bool(int(v)))
        # not needed, returns unicode in python2 and str in python3
        # self.dbh.text_factory =
        self.execute("pragma foreign_keys=1")
        self.logger.debug("database opened")

    def close(self):
        """
        Close the db connection, cleanup.
        """
        self.dbh.close()
        self.filename = None

    def commit(self):
        """
        Commit changes in the connection.
        """
        # Make sure the temporary entries in TMP_ADDSQL_CRIT are never committed
        self._cleanup_addsql_crit()  # just in case...
        self.dbh.commit()

    def rollback(self):
        """
        Rollback changes in the connection.
        """
        self.dbh.rollback()

    def execute(self, sql, parameters=None):
        """
        Perform a DB execute on a new cursor, additionally add error logging.

        Parameters:
            sql         the sql statement
            parameters  parameters for the statement (eiter tuple or dict)
                        (both handled down to curs.execute)
        Returns:
            new cursor object
        """
        self.logger.debug("execute: %s", sql)
        # parameters is either a dict or an iterable (list, tuple etc)
        # no mixed form allowed in sqlite3 API
        if parameters and isinstance(parameters, dict):
            self.logger.debug(
                "execute - PARAMS:\n %s",
                '\n '.join(['{}: {} ({})'.format(key, value, type(value))
                            for (key, value) in list(parameters.items())]))
        elif parameters:
            self.logger.debug(
                "execute - PARAMS:\n %s",
                '\n '.join(['{} ({})'.format(val, type(val))
                            for val in parameters]))
        curs = self.dbh.cursor()
        if parameters:
            curs.execute(sql, parameters)
        else:
            curs.execute(sql)
        return curs

    def executemany(self, sql, parameters):
        """
        Perform a DB execute on a new cursor, additionally add error logging.

        Parameters:
            sql         the sql statement
            parameters  parameters for the statement (eiter tuple or dict)
                        (both handled down to curs.execute)
        Returns:
            new cursor object
        """
        self.logger.debug("execute: %s", sql)
        # parameters is an iterable of iterables (list, tuple etc)
        if parameters:
            self.logger.debug(
                "execute - PARAMS:\n %s",
                '\n '.join(['{} ({})'.format(val, type(val))
                            for row in parameters[:100]
                            for val in row]))
        curs = self.dbh.cursor()
        curs.executemany(sql, parameters)
        return curs

    @staticmethod
    def _convert_row2dict(result, desc):
        """
        Converts a fetched row (tupel) to a dictionary (keys=colnames).

        Parameters:
            result    fetched result row (tupel)
            desc      cursor description
        """
        ret = DataObject()
        for i, item in enumerate(result):
            colnam = desc[i][0]
            ret[colnam] = item
        return ret

    @classmethod
    def _convert_rows2dictlist(cls, resultlist, desc):
        """
        Converts a fetched row (tupel) to a dictionary (keys=colnames).

        Parameters:
            resultlist  fetched result rows (list of tupel)
            desc        cursor description
        """
        ret = []
        for row in resultlist:
            ret.append(cls._convert_row2dict(row, desc))
        return ret

    @staticmethod
    def _hostvar_sql(varname):
        """
        Returns the sql part of a host variable name.
        """
        return ":{}".format(varname)

    @staticmethod
    def _hostvar_param(varname):
        """
        Returns the parameter dictionary key for a host variable.
        """
        return varname

    def _insert_addsql_crit(self, in_list):
        """
        Insert table entries for addsql_crit.
        Parameters:
            in_list    list of criteria values
        Returns:
            temp_id    id for the values in the temp table
        """
        if any([isinstance(elem[1], six.string_types) for elem in in_list]):
            valtype = 'VAL_CHR'
        elif any([isinstance(elem[1], float) for elem in in_list]):
            valtype = 'VAL_DBL'
        else:
            valtype = 'VAL_INT'
        temp_sql = dedent("""\
             insert into TMP_ADDSQL_CRIT (ID, TIME, {})
                 values (:ID, CURRENT_TIMESTAMP, :VAL)""".format(valtype))
        temp_id = uuid.uuid1().hex
        self.tmp_addsql_crit_tables.append(temp_id)
        temp_params = {'ID': temp_id}
        for elem in in_list:
            temp_params['VAL'] = elem[1]
            self.execute(temp_sql, temp_params)
        return "select {} from TMP_ADDSQL_CRIT where ID='{}'".format(
            valtype, temp_id)

    def _cleanup_addsql_crit(self):
        """
        Cleanup temp table entries from addsql_crit
        Parameters:
            None
        Returns:
            None
        """
        sql = "delete from TMP_ADDSQL_CRIT where ID=:ID"
        params = {}
        for id_ in self.tmp_addsql_crit_tables:
            params['ID'] = id_
            self.execute(sql, params)
        self.tmp_addsql_crit_tables = []



"""
ebas/io/base.py
$Id: check_metadata.py 2503 2020-10-14 16:59:06Z pe $

Base classes for (file) I/O.

TODO: fuctionality which should be implemented in check_metadata is mixed into
ebasmetadata (parsing/checking is _not_ separated). This means that those checks
are only performed for files read, but not for file objects constructed for
writiting.
"""

from .base import EbasFileBase, EbasMetadataInvalid, isvalid
import datetime
from nilutility.list_helper import all_none, all_equal
from ebas.domain.basic_domain_logic.time_period import estimate_period_code, \
    estimate_resolution_code, estimate_sample_duration_code, \
    period_code_seconds
from ebas.domain.masterdata.am import EbasMasterAM
from ebas.domain.masterdata.ax import EbasMasterAX
from ebas.domain.masterdata.ca import EbasMasterCA
from ebas.domain.masterdata.co import EbasMasterCO
from ebas.domain.masterdata.fm import EbasMasterFM
from ebas.domain.masterdata.pm import EbasMasterPM
from ebas.domain.masterdata.qm_qv import EbasMasterQM, EbasMasterQV
from ebas.domain.masterdata.ip import EbasMasterIP
from ebas.domain.masterdata.se import EbasMasterSE
from ebas.domain.masterdata.sm import EbasMasterSM
from ebas.domain.masterdata.pr import EbasMasterPR


class EbasFileCheckMetadata(EbasFileBase):  # pylint: disable=W0223
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Partial class for EbasFile (Base class for file objects).
    This part handles checking of the metadata.
    """

    def check_metadata(self, ignore_rescode=False, ignore_revdate=False,
                       ignore_parameter=False, ignore_sampleduration=False):
        """
        Performs a check of the metadata of the object.
        Parameter:
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
        TODO: currently, most of the metadata syntax check is done during
        parsing, this checks are not available here. This should be fixed...
        """
        self.harmonize_persons()
        self.check_interdependent_metadata(
            ignore_rescode=ignore_rescode, ignore_revdate=ignore_revdate,
            ignore_parameter=ignore_parameter,
            ignore_sampleduration=ignore_sampleduration)

    def harmonize_persons(self):
        """
        Harmonize additional metadata in Originator and Submitter if both are
        the same persons.
        Ref: http://jira.nilu.no/browse/EWB-432

        Parmaters:
            None
        Returns:
            None
        """
        passes = (('originator', 'submitter'), ('submitter', 'originator'))
        for cur in passes:
            for pers in self.metadata[cur[0]]:
                if set([key for key in pers.keys()
                        if pers[key] is not None]) == \
                   set(['PS_LAST_NAME', 'PS_FIRST_NAME']):
                    # person has honly first and last name, look for same person
                    # in the other category:
                    res = [x for x in self.metadata[cur[1]]
                           if x.PS_LAST_NAME == pers.PS_LAST_NAME and
                           x.PS_FIRST_NAME == pers.PS_FIRST_NAME]
                    if res and \
                       set([key for key in res[0].keys()
                            if res[0][key] is not None]) != \
                       set(['PS_LAST_NAME', 'PS_FIRST_NAME']):
                        pers.update(res[0])
                        self.warning(
                            '{} {}, {}: no additional metadata '
                            'specified; using metadata from the {} with the '
                            'same name.'
                            .format(cur[0].capitalize(), pers.PS_LAST_NAME,
                                    pers.PS_FIRST_NAME, cur[1]))

    def check_interdependent_metadata(
            self, ignore_rescode=False, ignore_revdate=False,
            ignore_parameter=False, ignore_sampleduration=False):
        """
        Checks metadata consistency within the file.
        Parameters:
            ignore_rescode    ignore inconsistencies in rescode (downgrade error
                              to warning)
            ignore_revdate    ignore inconsistencies in revdate (downgrade error
                              to warning)
            ignore_paramter   ignore illegal parameters
        Returns:
            None
        """
        self._check_interdep_license()
        self._check_interdep_station()
        self._check_interdep_periodcode()
        self._check_interdep_rescode(ignore_rescode=ignore_rescode)
        self._check_interdep_durationcode(
            ignore_sampleduration=ignore_sampleduration)
        self._check_interdep_revdate(ignore_revdate=ignore_revdate)
        self._check_interdep_qa()
        self._check_interdep_uncertainty()
        self._check_interdep_detectionlimit()
        self._check_interdep_upperrangelimit()
        self._check_interdep_instrument()
        self._check_interdep_inlet()
        self._check_interdep_analytical()
        self._check_interdep_sensortype()
        self._check_interdep_absorption_crosssection()
        self._check_interdep_calibrationscale()
        self._check_interdep_stdmethod()
        self._check_interdep_parameter(ignore_parameter=ignore_parameter)

        orig_filename = self.metadata.filename
        self.gen_filename()
        if self.metadata.filename != orig_filename:
            self.warning("File name: changed from '{}' to '{}'".format(
                orig_filename, self.metadata.filename))

    def _check_interdep_license(self):
        """
        Checks project and license consistency.
        Parameters:
            None
        Returns:
            None
        """
        master = EbasMasterPR()
        if not isvalid(self.metadata.license):
            # no need to check
            return
        if isvalid(self.metadata.projects):
            is_open = False
            check_closed = True
            for proj in self.metadata.projects:
                try:
                    pr_elem = master[proj]
                except KeyError:
                    # error is written alreadt in ansa_ames.parse...
                    # but if we have an illegal project, we will not check on
                    # closed projects...
                    check_closed = False
                if pr_elem['OPEN_ACCESS']:
                    is_open = True
                    break
            if is_open and self.metadata.license != \
                    'https://creativecommons.org/licenses/by/4.0/':
                self.error("Wrong license for open data: '{}'".format(
                    self.metadata.license))
            if not is_open and check_closed:
                self.error("Wrong license for restricted data: '{}'".format(
                    self.metadata.license))

    def _check_interdep_station(self):
        """
        Checks Station and Platform consistency.
        Parameters:
            None
        Returns:
            None
        """
        if isvalid(self.metadata.station_code) and \
           isvalid(self.metadata.platform_code) and \
           not self.metadata.platform_code.startswith(\
               self.metadata.station_code[0:6]):
            self.error('Station Code and Platform code must match in the first '
                       '6 characters')
        # station is (still?) file global, no need to check for each variable

    def _check_interdep_periodcode(self):
        """
        Check consistency of period code.
        Parameters:
            None
        Returns:
            None
        """
        if isvalid(self.metadata.period) and len(self.sample_times) > 0 and \
           self.sample_times[0][0] is not None and self.sample_times[-1][1]:
            # periode code can not be checked, if there are no data
            # (len(sample_times) == 0): make sure files without data report an
            # error anyway (see read())
            period = estimate_period_code(self.sample_times[0][0],
                                          self.sample_times[-1][1])
            if self.metadata.period != period:
                self.warning("Period code '{}' should be '{}' according to "
                             "data".format(self.metadata.period, period))
        # period code is (by definition) file global, no need to check for each
        # variable

    def _check_interdep_rescode(self, ignore_rescode=False):
        """
        Check consistency of resolution code. Check against the median of all
        start time differences in data.
        Parameters:
            ignore_rescode    ignore inconsistencies in rescode (downgrade error
                              to warning)
        Returns:
            None
        """
        if isvalid(self.metadata.resolution) and len(self.sample_times) > 1:
            res_cde = estimate_resolution_code(self.sample_times)
            res_sec = period_code_seconds(res_cde)
            if not ignore_rescode and \
               self.metadata.resolution != res_cde and \
               abs(period_code_seconds(self.metadata.resolution) - \
                   res_sec) > res_sec * 0.25:
                # > 25% off: ERROR
                self.error("Resolution code '{}' should be '{}' according "
                           "to data".format(self.metadata.resolution,
                                            res_cde))
            elif self.metadata.resolution != res_cde and \
               abs(period_code_seconds(self.metadata.resolution) - \
                   res_sec) > res_sec * 0.1:
                # > 10% off: WARNING
                self.warning("Resolution code '{}' should be '{}' "
                             "according to data"\
                             .format(self.metadata.resolution, res_cde))
        # resolution code is (by definition) file global, no need to check for
        # each variable

    def _check_interdep_durationcode(self, ignore_sampleduration=False):
        """
        Check consistency of duration code. Check against the median of all
        sample durations in the data.
        Parameters:
            None
        Returns:
            None
        """
        if isvalid(self.metadata.duration) and len(self.sample_times) > 0:
            # check sample duration against median duration in data
            dur_cde = estimate_sample_duration_code(self.sample_times)
            if dur_cde:
                if self.metadata.duration == dur_cde:
                    return
                dur_sec = period_code_seconds(dur_cde)
                diff = abs(period_code_seconds(self.metadata.duration) -
                           dur_sec)
                if ignore_sampleduration and diff > dur_sec * 0.25:
                    # > 20% off: ERROR
                    self.warning("Sample duration '{}' should be '{}' according "
                               "to data".format(self.metadata.duration,
                                                dur_cde))
                elif diff > dur_sec * 0.25:
                    # > 20% off: ERROR
                    self.error("Sample duration '{}' should be '{}' according "
                               "to data".format(self.metadata.duration,
                                                dur_cde))
                elif diff > dur_sec * 0.1:
                    # > 10% off: WARNING
                    self.warning("Sample duration '{}' should be '{}' "
                                 "according to data"\
                                 .format(self.metadata.duration, dur_cde))
            else:
                self.error(
                    "Sample duration could not be checked because of errors in "
                    "start/end times (see previous error messages)")
        # duration code is (by definition) file global, no need to check for
        # each variable

    def _check_interdep_revdate(self, ignore_revdate=False):
        """
        Check consistency of revision date.
        Parameters:
            ignore_revdate    ignore inconsistencies in revdate (downgrade error
                              to warning)
        Returns:
            None
        """
        if not isvalid(self.metadata.revdate):
            return

        now = datetime.datetime.utcnow()
        # check reported revision date against last sample end time
        if len(self.sample_times) > 0 and \
           self.sample_times[-1][1] is not None and \
           self.metadata.revdate < self.sample_times[-1][1]:
            # revision date can not be checked, if there are no data
            # (len(sample_times) == 0): make sure files without data report an
            # error anyway (see read())

            # check rev date > last sample
            # This is only true for observed data (not for model data).
            # Needs reconsideration when we accept regimes other than IMG
            if ignore_revdate:
                self.warning(
                    'Revision date ({}) should not be earlier than last sample '
                    'endtime ({})'.format(
                        self.metadata['revdate'].strftime("%Y%m%d%H%M%S"),
                        self.sample_times[-1][1].strftime("%Y%m%d%H%M%S")))
            else:
                self.error(
                    'Revision date ({}) must not be earlier than last sample '
                    'endtime ({})'.format(
                        self.metadata['revdate'].strftime("%Y%m%d%H%M%S"),
                        self.sample_times[-1][1].strftime("%Y%m%d%H%M%S")))
        if self.metadata.revdate >= now:
            self.error('Revision date ({}) must not be in the future'.\
                       format(self.metadata.revdate.strftime("%Y%m%d%H%M%S")))

        # check for all variables:
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            if 'revdate' in self.variables[i].metadata:
                if len(self.sample_times) > 0 and \
                   self.sample_times[-1][1] is not None and \
                   self.variables[i].metadata['revdate'] < \
                   self.sample_times[-1][1]:
                    # revision date can not be checked, if there are no data
                    # (len(sample_times) == 0): make sure files without data
                    # report an error anyway (see read())

                    # check rev date > last sample
                    # This is only true for observed data (not for model data).
                    # Needs reconsideration when we accept regimes other than
                    # IMG
                    self.error(
                        'Variable {}: Revision date ({}) must not be earlier '
                        'than last sample endtime ({})'.\
                        format(
                            file_vnum,
                            self.variables[i].metadata['revdate'].strftime(
                                "%Y%m%d%H%M%S"),
                            self.sample_times[-1][1].strftime("%Y%m%d%H%M%S")),
                        lnum=file_metalnum)
                if self.variables[i].metadata['revdate'] > \
                   self.metadata['revdate']:
                    # check variable specific revdate is not greater than
                    # revdate:
                    self.error(
                        'Variable {}: Revision date specified for variable '
                        '({}) must not be later than globally set Revision '
                        'date ({})'.\
                        format(
                            file_vnum,
                            self.variables[i].metadata['revdate'].strftime(
                                "%Y%m%d%H%M%S"),
                            self.metadata.revdate.strftime("%Y%m%d%H%M%S")),
                        lnum=file_metalnum)
                if self.variables[i].metadata['revdate'] >= now:
                    self.error(
                        'Variable {}: Revision date ({}) must '
                        'not be in the future'.format(
                            file_vnum,
                            self.variables[i].metadata['revdate'].strftime(
                                "%Y%m%d%H%M%S")),
                        lnum=file_metalnum)

    def _check_interdep_qa(self):
        """
        Check consistency of QA metadata
        Parameters:
            None
        Returns:
            None
        """
        qv_ = EbasMasterQV()
        qm_ = EbasMasterQM()
        for qanum in self.get_qa_numbers():
            qanum_txt = qanum if qanum != -1 else ""
            main_qa = self.get_qa_by_number(qanum, None)
            if main_qa is None:
                main_qa = {}
            main_ft_type = self.metadata.instr_type
            # only use main component name if it is a "real" component, and
            # not just an arbitrary name:
            if self._is_real_comp(self.metadata.comp_name):
                main_comp_name = self.metadata.comp_name
            else:
                main_comp_name = None
            # check global metadata:
            if isvalid(main_qa['qm_id']):
                if not qm_.new_reference_calibration(main_qa['qm_id']):
                    # check QV validity:
                    if isvalid(main_ft_type) and \
                       isvalid(main_comp_name):
                        try:
                            # QV masterdata lookup with additional data level:
                            # if QV is not defined in DB masterdata, lookup in
                            # exceptional masterdata (data level dependent,
                            # not meant to be imported)
                            qv_[(main_qa['qm_id'], main_ft_type,
                                 main_comp_name, self.metadata.datalevel)]
                        except KeyError:
                            self.error(
                                "QA{} measure ID/Instrument type/Component "
                                "combination '{}'/'{}'/'{}' is not defined"
                                .format(qanum_txt, main_qa['qm_id'], main_ft_type,
                                        main_comp_name))

                    qm_dict = qm_[main_qa['qm_id']]
                    # check QA date is within QM date start and date end:
                    if (qm_dict.QM_START and isvalid(main_qa['qa_date']) and \
                        main_qa['qa_date'] < qm_dict.QM_START) or \
                       (qm_dict.QM_END and isvalid(main_qa['qa_date']) and \
                        main_qa['qa_date'] > qm_dict.QM_END):
                        if qm_dict.QM_START and qm_dict.QM_END:
                            self.error(
                                "QA{} date must be between {} and {}".format(
                                    qanum_txt, qm_dict.QM_START, qm_dict.QM_END))
                        elif qm_dict.QM_START:
                            self.error(
                                "QA{} date must be at or after {}".format(
                                    qanum_txt, qm_dict.QM_START))
                        else:
                            self.error(
                                "QA{} date must be at or before {}".format(
                                    qanum_txt, qm_dict.QM_END))
                    # check QA document metadata are consistent with QM:
                    if qm_dict.QM_DOCUMENT_TITLE and main_qa['qa_doc_name'] and \
                       qm_dict.QM_DOCUMENT_TITLE != main_qa['qa_doc_name']:
                        self.error(
                            "QA{} document name must be'{}'".format(
                                qanum_txt, qm_dict.QM_DOCUMENT_TITLE))
                    if qm_dict.QM_DOCUMENT_DATE and main_qa['qa_doc_date'] and \
                       qm_dict.QM_DOCUMENT_DATE != main_qa['qa_doc_date']:
                        date_txt = qm_dict.QM_DOCUMENT_DATE.strftime("%Y%m%d") \
                            if qm_dict.QM_DOCUMENT_DATE.time() == \
                                datetime.time(0, 0) \
                            else qm_dict.QM_DOCUMENT_DATE.strftime(
                                "%Y%m%d$H%M%S")
                        self.error(
                            "QA{} document date must be '{}'".format(
                                qanum_txt, date_txt))
                    if qm_dict.QM_DOCUMENT_URL and main_qa['qa_doc_url'] and \
                       qm_dict.QM_DOCUMENT_URL != main_qa['qa_doc_url']:
                        self.error(
                            "QA{} document name must be '{}'".format(
                                qanum_txt, qm_dict.QM_DOCUMENT_URL))
                    # check qa bias unit: must be consistent within NNCOM
                    if isvalid(main_qa['qa_bias']) and \
                       main_qa['qa_bias'][1] != '%' and \
                       self.metadata.unit != main_qa['qa_bias'][1]:
                        self.error("Unit '{}' used for 'QA{} bias' is "
                            "inconsistent with 'Unit' ({})".format(
                                main_qa['qa_bias'][1], qanum_txt,
                                self.metadata.unit))
                    # check qa variability unit: must be consistent within NNCOM
                    if isvalid(main_qa['qa_variability']) and \
                       main_qa['qa_variability'][1] != '%' and \
                       self.metadata.unit != main_qa['qa_variability'][1]:
                        self.error("Unit '{}' used for 'QA{} variability' is "
                            "inconsistent with 'Unit' ({})".format(
                                main_qa['qa_variability'][1], qanum_txt,
                                self.metadata.unit))


            # check varaible specific metadata:
            for vnum in range(len(self.variables)):
                file_vnum, _, file_metalnum = self.msgref_vnum_lnum(vnum)
                qa_ = main_qa.copy()  # need a copy, because of update below
                # update with var metadata:
                var_qa = self.get_qa_by_number(qanum, vnum)
                if var_qa:
                    qa_.update(var_qa)
                if isinstance(qa_['qm_id'], EbasMetadataInvalid):
                    continue
                if self._is_empty_qa(qa_):
                    continue
                if qm_.new_reference_calibration(main_qa['qm_id']):
                    continue
                
                # check mandatory metadata
                mandatory = (
                    ('qm_id', 'QA{} measure ID'.format(qanum_txt)),
                    ('qa_date', 'QA{} date'.format(qanum_txt)))
                    #('qa_outcome', 'QA{} outcome'.format(qanum_txt)))
                for err in [key[1] for key in mandatory
                            if qa_[key[0]] is None]:
                    self.error(
                        "Variable {}: {} is mandatory".format(file_vnum,
                                                              err),
                        lnum=file_metalnum)

                ft_type = self.get_meta_for_var(vnum, "instr_type")
                comp_name = self.get_meta_for_var(vnum, "comp_name")
                # check QV validity:
                if isvalid(ft_type) and isvalid(comp_name) and \
                   (qa_['qm_id'] != main_qa['qm_id'] or \
                    ft_type != main_ft_type or \
                    comp_name != main_comp_name):
                    try:
                        # IP masterdata lookup with additional data level:
                        # if IP is not defined in DB masterdata, lookup in
                        # exceptional masterdata (data level dependent, not
                        # meant to be imported)
                        qv_[(qa_['qm_id'], ft_type, comp_name,
                             self.metadata.datalevel)]
                    except KeyError:
                        self.error(
                            "Variable {}: QA{} measure ID/Instrument type/"
                            "Component combination '{}'/'{}'/'{}' is not "
                            "defined"
                            .format(file_vnum, qanum_txt, qa_['qm_id'],
                                    ft_type, comp_name),
                            lnum=file_metalnum)
                if isvalid(qa_['qm_id']):
                    qm_dict = qm_[qa_['qm_id']]
                    # check QA date is within QM date start and date end:
                    if qa_['qm_id'] != main_qa['qm_id'] or \
                       qa_['qa_date'] != main_qa['qa_date']:
                        if (qm_dict.QM_START and qa_['qa_date'] and \
                            qa_['qa_date'] < qm_dict.QM_START) or \
                           (qm_dict.QM_END and qa_['qa_date'] and \
                            qa_['qa_date'] > qm_dict.QM_END):
                            if qm_dict.QM_START and qm_dict.QM_END:
                                self.error(
                                    "Variable {}: QA{} date must be between {} and "
                                    "{}".format(file_vnum, qanum_txt,
                                                qm_dict.QM_START, qm_dict.QM_END),
                                    lnum=file_metalnum)
                            elif qm_dict.QM_START:
                                self.error(
                                    "Variable {}: QA{} date must be at or after {}"
                                    .format(file_vnum, qanum_txt, qm_dict.QM_START),
                                    lnum=file_metalnum)
                            else:
                                self.error(
                                    "Variable {}: QA{} date must be at or before {}"
                                    .format(file_vnum, qanum_txt, qm_dict.QM_END),
                                    lnum=file_metalnum)
                    # check QA document metadata are consistent with QM:
                    if qa_['qm_id'] != main_qa['qm_id'] or \
                       qa_['qa_doc_name'] != main_qa['qa_doc_name']:
                        if qm_dict.QM_DOCUMENT_TITLE and qa_['qa_doc_name'] and \
                           qm_dict.QM_DOCUMENT_TITLE != qa_['qa_doc_name']:
                            self.error(
                                "Variable {}: QA{} document name must be'{}'"
                                .format(file_vnum, qanum_txt,
                                        qm_dict.QM_DOCUMENT_TITLE),
                                lnum=file_metalnum)
                    if qa_['qm_id'] != main_qa['qm_id'] or \
                       qa_['qa_doc_date'] != main_qa['qa_doc_date']:
                        if qm_dict.QM_DOCUMENT_DATE and qa_['qa_doc_date'] and \
                           qm_dict.QM_DOCUMENT_DATE != qa_['qa_doc_date']:
                            date_txt = qm_dict.QM_DOCUMENT_DATE.strftime("%Y%m%d") \
                                if qm_dict.QM_DOCUMENT_DATE.time() == \
                                    datetime.time(0, 0) \
                                else qm_dict.QM_DOCUMENT_DATE.strftime(
                                    "%Y%m%d$H%M%S")
                            self.error(
                                "Variable {}: QA{} document date must be '{}'"
                                .format(file_vnum, qanum_txt, date_txt),
                                lnum=file_metalnum)
                    if qa_['qm_id'] != main_qa['qm_id'] or \
                       qa_['qa_doc_url'] != main_qa['qa_doc_url']:
                        if qm_dict.QM_DOCUMENT_URL and qa_['qa_doc_url'] and \
                           qm_dict.QM_DOCUMENT_URL != qa_['qa_doc_url']:
                            self.error(
                                "Variable {}: QA{} document name must be '{}'"
                                .format(file_vnum, qanum_txt,
                                        qm_dict.QM_DOCUMENT_URL),
                                lnum=file_metalnum)

                    unit = self.get_meta_for_var(vnum, 'unit')
                    # check qa bias unit: must be consistent for each variable
                    # check always, even if qm_id and qm_variablility is the
                    # same as in NCOM (check unit against var's unit!)
                    if isvalid(qa_['qa_bias']) and \
                        qa_['qa_bias'][1] != '%' and \
                        qa_['qa_bias'][1] != unit:
                        self.error(
                            "Variable {}: Unit '{}' used for 'QA{} bias' "
                            "is inconsistent with variable's unit ('{}')"
                                .format(file_vnum, qa_['qa_bias'][1],
                                        qanum_txt, unit))
                    # check qa variability unit: must be consistent for each
                    # variable
                    # check always, even if qm_id and qm_variablility is the
                    # same as in NCOM (check unit against var's unit!)
                    if isvalid(qa_['qa_variability']) and \
                        qa_['qa_variability'][1] != '%' and \
                        qa_['qa_variability'][1] != unit:
                        self.error(
                            "Variable {}: Unit '{}' used for 'QA{} variability' "
                            "is inconsistent with variable's unit ('{}')"
                                .format(file_vnum, qa_['qa_variability'][1],
                                        qanum_txt, unit))


    def _check_interdep_uncertainty(self):
        """
        Check consistency of uncertainty.
        Parameters:
            None
        Returns:
            None
        """
        # check uncertainty/unit: must be consistent within NNCOM
        if isvalid(self.metadata.uncertainty) and \
           self.metadata.uncertainty[1] != '%' and \
           self.metadata.unit != self.metadata.uncertainty[1]:
            self.error("Unit '{}' ".format(self.metadata.uncertainty[1]) +\
                "used for 'Measurement uncertainty' is inconsistent with " +\
                "'Unit' ({})".format(self.metadata.unit))
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            unit = self.get_meta_for_var(i, 'unit')

            # special: ignore NNCOM uncertainty for variables where unit does
            # not match (relative uncertainties [%] match ALL variables)
            if 'uncertainty' not in self.variables[i].metadata and \
               isvalid(self.metadata.uncertainty) and \
               self.metadata.uncertainty[1] != '%' and \
               self.metadata.uncertainty[1] != unit:
                self.warning(
                    "Variable {}: unit '{}' used for 'Measurement uncertainty' "
                    "is inconsistent with variable's unit ('{}'); "
                    "uncertainty definition will not be applied".format(
                        file_vnum,
                        self.metadata.uncertainty[1], unit),
                    lnum=file_metalnum)
                self.variables[i].metadata['uncertainty'] = None
                if 'uncertainty_desc' not in self.variables[i].metadata:
                    self.variables[i].metadata['uncertainty_desc'] = None
                    self.warning(
                        "Variable {}: 'Measurement uncertainty expl.' will not "
                        "be applied for the same reason".format(
                            file_vnum),
                        lnum=file_metalnum)

            # check uncertainty/unit: must be consistent for each variable
            uncert = self.get_meta_for_var(i, 'uncertainty')
            if uncert and uncert[1] != '%' and uncert[1] != unit:
                self.error(
                    "Variable {}: unit '{}' used for 'Measurement uncertainty' "
                    "is inconsistent with variable's unit ('{}')".format(
                        file_vnum, uncert[1], unit),
                    lnum=file_metalnum)

    def _check_interdep_detectionlimit(self):
        """
        Check consistency of detection limit.
        Parameters:
            None
        Returns:
            None
        """
        # check detection limit/unit: must be consistent within NNCOM
        if isvalid(self.metadata.detection_limit) and \
           self.metadata.unit != self.metadata.detection_limit[1]:
            self.error("Unit '{}' ".format(self.metadata.detection_limit[1]) +\
                "used for 'Detection limit' is inconsistent with " +\
                "'Unit' ('{}')".format(self.metadata.unit))
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            unit = self.get_meta_for_var(i, 'unit')

            # special: ignore NNCOM detection limit for variables where unit
            # does not match
            if 'detection_limit' not in self.variables[i].metadata and \
               isvalid(self.metadata.detection_limit) and \
               self.metadata.detection_limit[1] != unit:
                self.warning(
                    "Variable {}: unit '{}' used for 'Detection limit' is "
                    "inconsistent with variable's unit ('{}'); "
                    "detection limit definition will not be applied".format(
                        file_vnum,
                        self.metadata.detection_limit[1], unit),
                    lnum=file_metalnum)
                self.variables[i].metadata['detection_limit'] = None
                if 'detection_limit_desc' not in self.variables[i].metadata:
                    self.variables[i].metadata['detection_limit_desc'] = None
                    self.warning(
                        "Variable {}: 'Detection limit expl.' will not be "
                        "applied for the same reason".format(
                            file_vnum),
                        lnum=file_metalnum)
            # check detection limit/unit: must be consistent for each variable
            detect = self.get_meta_for_var(i, 'detection_limit')
            if isvalid(detect) and detect[1] != unit:
                self.error(
                    "Variable {}: unit '{}' used for 'Detection limit' is "
                    "inconsistent with variable's unit ('{}')".format(
                        file_vnum, detect[1], unit),
                    lnum=file_metalnum)

    def _check_interdep_upperrangelimit(self):
        """
        Check consistency of upper range limit. Additionally check consistency
        of upper range limit and detection limit.
        Parameters:
            None
        Returns:
            None
        """
        # check upper range limit/unit: must be consistent within NNCOM
        if isvalid(self.metadata.upper_range_limit) and \
           self.metadata.unit != self.metadata.upper_range_limit[1]:
            self.error(
                "Unit '{}' ".format(self.metadata.upper_range_limit[1]) +\
                "used for 'Upper range limit' is inconsistent with " +\
                "'Unit' ('{}')".format(self.metadata.unit))
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            unit = self.get_meta_for_var(i, 'unit')

            # special: ignore NNCOM upper range limit for variables where unit
            # does not match
            if 'upper_range_limit' not in self.variables[i].metadata and \
               isvalid(self.metadata.upper_range_limit) and \
               self.metadata.upper_range_limit[1] != unit:
                self.warning(
                    "Variable {}: unit '{}' used for 'Upper range limit' is "
                    "inconsistent with variable's unit ('{}'); upper range "
                    "limit definition will not be applied".format(
                        file_vnum,
                        self.metadata.upper_range_limit[1], unit),
                    lnum=file_metalnum)
                self.variables[i].metadata['upper_range_limit'] = None

            # check upper range limit/unit: must be consistent for each variable
            upper = self.get_meta_for_var(i, 'upper_range_limit')
            if upper and upper[1] != unit:
                self.error(
                    "Variable {}: unit '{}' used for 'Upper range limit' is "
                    "inconsistent with variable's unit ('{}')".format(
                        file_vnum, upper[1], unit),
                    lnum=file_metalnum)

            # check detection limit lower then upper range limit
            detect = self.get_meta_for_var(i, 'detection_limit')
            upper = self.get_meta_for_var(i, 'upper_range_limit')
            if detect and upper and detect[0] >= upper[0]:
                self.error(
                    "Variable {}: detection limit ({} {}) must be lower then "
                    "upper range limit ({} {})".format(
                        file_vnum,
                        detect[0], detect[1], upper[0], upper[1]),
                    lnum=file_metalnum)

    def _check_interdep_instrument(self):
        """
        Check consistency of instrument metadata.
        Parameters:
            None
        Returns:
            None
        """
        # instrument metadata must be only consistent within NNCOM if all three
        # are set
        allset = False
        if isvalid(self.metadata.instr_type) and \
           isvalid(self.metadata.instr_manufacturer) and \
           isvalid(self.metadata.instr_model):
            # check global metadata consistency only when ALL THREE are set!
            allset = True
            try:
                EbasMasterFM()[
                    (self.metadata.instr_type, self.metadata.instr_manufacturer,
                     self.metadata.instr_model)]
            except KeyError:
                self.error(
                    "Instrument type/manufacturer/model "
                    "combination '{}'/'{}'/'{}' is not defined".format(
                        self.metadata.instr_type,
                        self.metadata.instr_manufacturer,
                        self.metadata.instr_model))

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            instr_type = self.get_meta_for_var(i, 'instr_type')
            fm_man = self.get_meta_for_var(i, 'instr_manufacturer')
            fm_mod = self.get_meta_for_var(i, 'instr_model')
            if fm_man is not None or fm_mod is not None:
                # model or manufacturer is set for this variable: check
                if allset and \
                   'instr_type' not in self.variables[i].metadata and \
                   'instr_manufacturer' not in self.variables[i].metadata and \
                   'instr_model' not in self.variables[i].metadata:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                # check if any is not set or set with an illegal value:
                dependencies = []
                if not isvalid(instr_type):
                    dependencies.append('instrument type')
                if not isvalid(fm_man):
                    dependencies.append('instrument manufacturer')
                if not isvalid(fm_mod):
                    dependencies.append('instrument model')
                if dependencies:
                    self.error(
                        "Variable {}: Instrument type/manufacturer/model "
                        "combination could not be checked because {} is not "
                        "specified correctly".format(
                            file_vnum, " and ".join(dependencies)),
                        lnum=file_metalnum)
                else:
                    try:
                        EbasMasterFM()[(instr_type, fm_man, fm_mod)]
                    except KeyError:
                        self.error(
                            "Variable {}: Instrument type/manufacturer/model "
                            "combination '{}'/'{}'/'{}' is not defined".format(
                                file_vnum, instr_type, fm_man, fm_mod),
                            lnum=file_metalnum)

    def _check_interdep_inlet(self):
        """
        Check consistency of Inlet tube outer and inner diameter.
        Parameters:
            None
        Returns:
            None
        """
        if isvalid(self.metadata.inlet_tube_outerD) and \
           isvalid(self.metadata.inlet_tube_innerD) and \
            self.metadata.inlet_tube_outerD <= self.metadata.inlet_tube_innerD:
            self.error(
                "Inlet tube outer diameter ({} mm) is not greater than "
                "inlet tube inner diameter ({} mm)".format(
                    self.metadata.inlet_tube_outerD,
                    self.metadata.inlet_tube_innerD))
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            outer = self.get_meta_for_var(i, 'inlet_tube_outerD')
            inner = self.get_meta_for_var(i, 'inlet_tube_innerD')
            if outer is not None and inner is not None:
                # innerD and outerD is set for this variable: check
                if 'inlet_tube_outerD' not in self.variables[i].metadata and \
                   'inlet_tube_innerD' not in self.variables[i].metadata:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                # check if any is not set or set with an illegal value:
                if outer <= inner:
                    self.error(
                        "Variable {}: Inlet tube outer diameter ({} mm) is not "
                        "greater than inlet tube inner diameter ({} mm)".format(
                            file_vnum, outer, inner), lnum=file_metalnum)

    def _check_interdep_analytical(self):
        """
        Check consistency of analytical metadata.
        Parameters:
            None
        Returns:
            None
        """
        # instrument metadata must be only consistent within NNCOM if all three
        # are set
        allset = False
        if isvalid(self.metadata.ana_technique) and \
           isvalid(self.metadata.ana_instr_manufacturer) and \
           isvalid(self.metadata.ana_instr_model):
            # check global metadata consistency only when ALL THREE are set!
            allset = True
            try:
                EbasMasterAM()[(self.metadata.ana_technique,
                                self.metadata.ana_instr_manufacturer,
                                self.metadata.ana_instr_model)]
            except KeyError:
                self.error(
                    "Analytical instrument technique/manufacturer/model "
                    "combination '{}'/'{}'/'{}' is not defined".format(
                        self.metadata.ana_technique,
                        self.metadata.ana_instr_manufacturer,
                        self.metadata.ana_instr_model))

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            ana_technique = self.get_meta_for_var(i, 'ana_technique')
            am_man = self.get_meta_for_var(i, 'ana_instr_manufacturer')
            am_mod = self.get_meta_for_var(i, 'ana_instr_model')

            if am_man is not None or am_mod is not None:
                # model or manufacturer is set for this variable: check
                if allset and \
                   'ana_technique' not in self.variables[i].metadata and \
                   'am_man' not in self.variables[i].metadata and \
                   'am_mod' not in self.variables[i].metadata:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                # check if any is not set or set with an illegal value:
                dependencies = []
                if not isvalid(ana_technique):
                    dependencies.append('analytical measurement technique')
                if not isvalid(am_man):
                    dependencies.append('analytical instrument manufacturer')
                if not isvalid(am_mod):
                    dependencies.append('analytical instrument model')
                if dependencies:
                    self.error(
                        "Variable {}: Analytical measurement technique/"
                        "manufacturer/model combination could not be checked "
                        "because {} is not specified correctly".format(
                            file_vnum, " and ".join(dependencies)),
                        lnum=file_metalnum)
                else:
                    try:
                        EbasMasterAM()[(ana_technique, am_man, am_mod)]
                    except KeyError:
                        self.error(
                            "Variable {}: Analytical measurement technique/"
                            "manufacturer/model combination '{}'/'{}'/'{}' is "
                            "not defined".format(
                                file_vnum, ana_technique, am_man, am_mod),
                            lnum=file_metalnum)

    def _check_interdep_sensortype(self):
        """
        Check consistency of sensor type and instrument type / component.
        Parameters:
            None
        Returns:
            None
        """
        # metadata must be only consistent within NNCOM if both are set.
        allset = False
        instr_type = self.metadata.instr_type
        sensor_type = self.metadata.sensor_type
        # only use main component name if it is a "real" component, and
        # not just an arbitrary name:
        if self._is_real_comp(self.metadata.comp_name):
            comp_name = self.metadata.comp_name
        else:
            comp_name = None
        if isvalid(instr_type) and isvalid(comp_name) and isvalid(sensor_type):
            allset = True
            try:
                _ = EbasMasterSE()[(instr_type, comp_name, sensor_type)]
            except KeyError:
                self.error(
                    "Instrument type/Component/Sensor type combination "
                    "'{}'/'{}'/'{}' is not defined").format(
                        instr_type, comp_name, sensor_type)

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            instr_type = self.get_meta_for_var(i, 'instr_type')
            comp_name = self.get_meta_for_var(i, 'comp_name')
            sensor_type = self.get_meta_for_var(i, 'sensor_type')
            if isvalid(instr_type) and isvalid(comp_name) and \
               isvalid(sensor_type):
                if allset and \
                   'instr_type' not in self.variables[i].metadata and \
                   comp_name == self.get_meta_for_var(i, 'comp_name') and \
                   'sensor_type' not in self.variables[i].metadata:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                try:
                    _ = EbasMasterSE()[(instr_type, comp_name, sensor_type)]
                except KeyError:
                    self.error(
                        "Variable {}: Instrument type/Component/Sensor type "
                        "combination '{}'/'{}'/'{}' is not defined".format(
                            file_vnum, instr_type, comp_name, sensor_type),
                        lnum=file_metalnum)

    def _check_interdep_absorption_crosssection(self):
        """
        Check consistency of absorption cross section and component.
        Parameters:
            None
        Returns:
            None
        """
        # metadata must be only consistent within NNCOM if both are set.
        allset = False
        main_abs_cross_section = self.metadata.abs_cross_section
        # only use main component name if it is a "real" component, and
        # not just an arbitrary name:
        if self._is_real_comp(self.metadata.comp_name):
            main_comp_name = self.metadata.comp_name
        else:
            main_comp_name = None
        if isvalid(main_comp_name) and isvalid(main_abs_cross_section):
            allset = True
            try:
                _ = EbasMasterAX()[(main_comp_name, main_abs_cross_section)]
            except KeyError:
                self.error(
                    "Component/Absorption cross section "
                    "'{}'/'{}' is not defined").format(
                        main_comp_name, main_abs_cross_section)

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            comp_name = self.get_meta_for_var(i, 'comp_name')
            abs_cross_section = self.get_meta_for_var(i, 'abs_cross_section')
            if isvalid(comp_name) and isvalid(abs_cross_section):
                if allset and \
                   comp_name == main_comp_name and \
                   abs_cross_section == main_abs_cross_section:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                try:
                    _ = EbasMasterAX()[(comp_name, abs_cross_section)]
                except KeyError:
                    self.error(
                        "Variable {}: Component/Absorption cross section "
                        "combination '{}'/'{}' is not defined".format(
                            file_vnum, comp_name, abs_cross_section),
                        lnum=file_metalnum)

    def _check_interdep_calibrationscale(self):
        """
        Check consistency of sensor type and instrument type / component.
        Parameters:
            None
        Returns:
            None
        """
        # metadata must be only consistent within NNCOM if both are set.
        allset = False
        main_cal_scale = self.metadata.cal_scale
        # only use main component name if it is a "real" component, and
        # not just an arbitrary name:
        if self._is_real_comp(self.metadata.comp_name):
            main_comp_name = self.metadata.comp_name
        else:
            main_comp_name = None
        if isvalid(main_comp_name) and isvalid(main_cal_scale):
            allset = True
            try:
                _ = EbasMasterCA()[(main_comp_name, main_cal_scale,
                                    self.metadata.datalevel)]
            except KeyError:
                self.error(
                    "Component/Calibration scale combination '{}'/'{}' is not "
                    "defined".format(main_comp_name, main_cal_scale))

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            comp_name = self.get_meta_for_var(i, 'comp_name')
            cal_scale = self.get_meta_for_var(i, 'cal_scale')
            if isvalid(comp_name) and isvalid(cal_scale):
                if allset and \
                   comp_name == main_comp_name and \
                   cal_scale == main_cal_scale:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                try:
                    _ = EbasMasterCA()[(comp_name, cal_scale,
                                        self.metadata.datalevel)]
                except KeyError:
                    self.error(
                        "Variable {}: Component/Calibration scale combination "
                        "'{}'/'{}' is not defined".format(
                            file_vnum, comp_name, cal_scale),
                        lnum=file_metalnum)

    def _check_interdep_stdmethod(self):
        """
        Check consistency of standard method and instrument type.
        Parameters:
            None
        Returns:
            None
        """
        # metadata must be only consistent within NNCOM if both are set.
        allset = False
        instr_type = self.metadata.instr_type
        std_method = self.metadata.std_method
        if isvalid(instr_type) and isvalid(std_method):
            allset = True
            try:
                _ = EbasMasterSM()[(instr_type, std_method)]
            except KeyError:
                self.error(
                    "Instrument type/Standard method combination '{}'/'{}' is "
                    "not defined".format(instr_type, std_method))

        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
            instr_type = self.get_meta_for_var(i, 'instr_type')
            std_method = self.get_meta_for_var(i, 'std_method')
            if isvalid(instr_type) and isvalid(std_method):
                if allset and \
                   'instr_type' not in self.variables[i].metadata and \
                   'std_method' not in self.variables[i].metadata:
                    # all are set in NCOM and none is set per variable:
                    # no need to check
                    continue
                try:
                    _ = EbasMasterSM()[(instr_type, std_method)]
                except KeyError:
                    self.error(
                        "Variable {}: Instrument type/Standard method "
                        "combination '{}'/'{}' is not defined".format(
                            file_vnum, instr_type, std_method),
                        lnum=file_metalnum)

    def _check_interdep_parameter(self, ignore_parameter=False):
        """
        Check consistency of parameter.
           1) validity of triple (regime, matrix, component)
           2) consistency of unit
           3) validity of the parameter for the instrument
        Parameters:
            ignore_parameter  ignore inconsistencies in parameters (downgrade
                              error to warning)
        Returns:
            None
        """
        # parameter metadata must be only consistent within NNCOM if all three
        # are set and component name is equal one of the vaiable's component
        # (this to allow for free component names: like VOC, inorganics, ...

        def check_param_unit(file_lnum, file_vnum, param, unit):
            """
            Helper for checking the unit is correct for parameter.
            """
            if param['PM_UNIT'] != unit:
                vtxt = ''
                if file_vnum:
                    vtxt = "Variable {}: ".format(file_vnum)
                msg = (
                    "{}Regime/Matrix/Component combination '{}'/'{}'/'{}': "
                    "unit '{}' not allowed, should be '{}'".format(
                        vtxt, param['RE_REGIME_CODE'], param['MA_MATRIX_NAME'],
                        param['CO_COMP_NAME'], unit, param['PM_UNIT']))
                if ignore_parameter:
                    self.warning(msg, lnum=file_lnum)
                else:
                    self.error(msg, lnum=file_lnum)

        def check_param_ip(file_lnum, file_vnum, instr_type, regime, matrix,
                           comp_name):
            """
            Helper for checking the intrument-type/parameter validity
            """
            try:
                # IP masterdata lookup with additional data level:
                # if IP is not defined in DB masterdata, lookup in exceptional
                # masterdata (data level dependent, not meant to be imported)
                EbasMasterIP()[
                    (instr_type, regime, matrix, comp_name,
                     self.metadata.datalevel)]
            except KeyError:
                vtxt = ''
                if file_vnum:
                    vtxt = "Variable {}: ".format(file_vnum)
                msg = (
                    "{}Regime/Matrix/Component combination '{}'/'{}'/'{}' "
                    "is not allowed for instrument type {}".format(
                        vtxt, regime, matrix, comp_name,
                        instr_type))
                if ignore_parameter:
                    self.warning(msg, lnum=file_lnum)
                else:
                    self.error(msg, lnum=file_lnum)

        regime = self.metadata.regime
        matrix = self.metadata.matrix
        comp_name = self.metadata.comp_name
        param = None
        instr_type = self.metadata.instr_type
        statistics = self.metadata.statistics
        unit = self.metadata.unit

        if isvalid(regime) and isvalid(matrix) and isvalid(comp_name) and \
           comp_name in [self.get_meta_for_var(i, 'comp_name')
                         for i in range(len(self.variables))]:
            # check global metadata consistency only when ALL THREE are set!
            try:
                # pm masterdata lookup with additional data level:
                # if pm is not defined in DB masterdata, lookup in exceptional
                # masterdata (data level dependent, not meant to be imported)
                param = EbasMasterPM()[(regime, matrix, comp_name, statistics,
                                        self.metadata.datalevel)]
            except KeyError:
                msg = (
                    "Regime/Matrix/Component combination '{}'/'{}'/'{}' is "
                    "not defined").format(regime, matrix, comp_name)
                if ignore_parameter:
                    self.warning(msg)
                else:
                    self.error(msg)
        if param and isvalid(unit):
            # check unit consistency
            check_param_unit(None, None, param, unit)
        if param and isvalid(instr_type):
            check_param_ip(None, None, instr_type, regime, matrix, comp_name)

        # do the checks for all variables:
        for i in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)

            # check paramter existence:
            v_regime = self.get_meta_for_var(i, 'regime')
            v_matrix = self.get_meta_for_var(i, 'matrix')
            v_comp_name = self.get_meta_for_var(i, 'comp_name')
            v_param = None
            v_unit = self.get_meta_for_var(i, 'unit')
            v_instr_type = self.get_meta_for_var(i, 'instr_type')
            v_statistics = self.get_meta_for_var(i, 'statistics')

            dependencies = []
            if not isvalid(v_regime):
                dependencies.append('Regime code')
            if not isvalid(v_matrix):
                dependencies.append('Matrix name')
            if not isvalid(v_comp_name):
                dependencies.append('Component name')
            if dependencies:
                msg = (
                    "Variable {}: Regime/Matrix/Component combination could "
                    "not be checked because {} is not specified correctly"
                    .format(file_vnum, " and ".join(dependencies)))
                if ignore_parameter:
                    self.warning(msg, lnum=file_metalnum)
                else:
                    self.error(msg, lnum=file_metalnum)
            else:
                if (v_regime, v_matrix, v_comp_name) == (regime, matrix,
                                                         comp_name):
                    v_param = param
                    if v_param and isvalid(v_unit) and v_unit != unit:
                        # the parameter triple is identical to global
                        # metadata, if the unit is identical as well, skip
                        # this check
                        check_param_unit(file_metalnum, file_vnum, v_param,
                                         v_unit)
                    if v_param and isvalid(v_instr_type) and \
                       v_instr_type != instr_type:
                        # the parameter triple is identical to global
                        # metadata, if the isntr_type is identical as well,
                        # skip this check
                        check_param_ip(file_metalnum, file_vnum, v_instr_type,
                                       v_regime, v_matrix, v_comp_name)
                else:
                    try:
                        # pm masterdata lookup with additional data level:
                        # if pm is not defined in DB masterdata, lookup in exceptional
                        # masterdata (data level dependent, not meant to be imported)
                        v_param = EbasMasterPM()[
                            (v_regime, v_matrix, v_comp_name, v_statistics,
                             self.metadata.datalevel)]
                    except KeyError:
                        msg = (
                            "Variable {}: Regime/Matrix/Component "
                            "combination '{}'/'{}'/'{}' is not defined"
                            .format(file_vnum, v_regime, v_matrix,
                                    v_comp_name))
                        if ignore_parameter:
                            self.warning(msg, lnum=file_metalnum)
                        else:
                            self.error(msg, lnum=file_metalnum)
                    if v_param and isvalid(v_unit):
                        # here, the parameter triple is already different, so
                        # check unit validity for this variable!
                        check_param_unit(file_metalnum, file_vnum, v_param,
                                         v_unit)
                    if v_param and isvalid(v_instr_type):
                        # here, the parameter triple is already different, so
                        # check instr_type validity for this variable!
                        check_param_ip(file_metalnum, file_vnum, v_instr_type,
                                       v_regime, v_matrix, v_comp_name)

    @staticmethod
    def _is_real_comp(co_comp_name):
        """
        Check if the component name is a "real" component name (exists in 
        CO_COMPONENT table). or a fake name (in the main header, like "GHG" etc.
        Parameters:
            co_comp_name    compoent name
        Returns:
            True if real name, False when fake name
        """
        co_ = EbasMasterCO()
        try:
            co_[co_comp_name].CO_COMP_NAME
        except KeyError:
            return False
        return True



"""
ebas/io/nasa_ames/write.py
$Id: write.py 2769 2021-12-08 22:35:41Z pe $

output functionality for EBAS NASA Ames module
"""

import re
import sys
import os
import textwrap
from nilutility.string_helper import list_joiner
from .base import EbasNasaAmesBase
from ..base import FLAGS_ONE_OR_ALL, FLAGS_NONE
from fileformats.NasaAmes1001 import NasaAmes1001


class NasaAmesPartialWrite(EbasNasaAmesBase):  # pylint: disable=R0901, W0223
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: '_read' is abstract in class 'EbasFileRead' but is not overridden
    # This class is also abstract.
    """
    Nasa Ames I/O object.

    This is one part of the partial class NasaAmes (output functionality).
    """

    def write(self, createfiles=False, destdir=None,  # pylint: disable=R0913
              xmlwrap=False, fileobj=None, flags=FLAGS_ONE_OR_ALL,
              datadef='EBAS_1.1', metadata_options=0, suppress=0):
        # R0913 Too many arguments
        """
        Writes the I/O object to a NASA Ames file (or stdout)
        Parameters:
            createfiles   write to file? else write to stdout
                          (special case for fileobj)
            destdir       destination directory path for files
            xmlwrap       wrap output in xml container
            fileobj       write to filehandle
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
            datadef       ebas data definition (EBAS_1 or EBAS_1.1)
            metadata_options
                          options for output, bitfield:
                          currently only EBAS_IOMETADATA_OPTION_SETKEY
            suppress      suppress selected consolidations (bitfield):
                              SUPPRESS_SORT_VARIABLES
                              SUPPRESS_METADATA_OCCURRENCE
        Returns:
            None
        """
        fil = self.prepare_write(
            True, createfiles=createfiles, destdir=destdir, xmlwrap=xmlwrap,
            fileobj=fileobj, flags=flags, datadef=datadef,
            metadata_options=metadata_options, suppress=suppress)

        self.nasa1001 = NasaAmes1001()
        self.set_header(flags=flags)
        self.set_data(flags=flags)
        self.nasa1001.write(fil)
        self.finish_write(createfiles, xmlwrap)
        if createfiles:
            fil.close()
        self.write_indexdb()

    def set_header(self, flags=FLAGS_ONE_OR_ALL):
        """
        Gnereate the header block of the NASA Ames file.
        Sets directly the NasaAmes1001 attributes.
        Parameters:
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
        Returns
            None
        """
        self.nasa1001.ONAME = self.gen_oname_str()
        self.nasa1001.ORG = self.gen_org_str()
        self.nasa1001.SNAME = self.gen_sname_str()
        self.nasa1001.MNAME = self.gen_mname_str()
        self.nasa1001.IVOL = 1
        self.nasa1001.NVOL = 1
        self.nasa1001.DATE = self.metadata.reference_date
        self.nasa1001.RDATE = self.metadata.revdate
        self.nasa1001.XNAME = "days from file reference point"
        # VSCAL defaults to 1 in NasaAmes1001
        self.nasa1001.VNAME = list(self.gen_vname_str())
        self.nasa1001.NCOML = list(self.gen_nncom_str())
        if flags == FLAGS_NONE:
            self.nasa1001.SCOML = textwrap.dedent("""\
            !!! WARNING !!!
            This file has been extracted from EBAS without flag information.
            Data without flag information are easier to process by non expert
            users, but part of the information is lost!
            Changes on extract without flag information:
                - values flagged with at least one flag from the class of
                  invalid flags are changed to MISSING VALUE
                - values flagged with flag 781 ("Value below detection limit,
                  data element contains detection limit") are divided by 2""").\
            split('\n')

    def gen_oname_str(self):
        """
        Generates output string for ONAME line.
        Parameter:
            None
        Returns:
            ONAME line (str)
        """
        return self._gen_persname_str(self.metadata.originator)

    def gen_sname_str(self):
        """
        Generates output string for SNAME line.
        Parameter:
            None
        Returns:
            SNAME line (str)
        """
        return self._gen_persname_str(self.metadata.submitter)

    @staticmethod
    def _gen_persname_str(perslist):
        """
        Creates output string for SNAME or ONAME line.
        Parameters:
            perslist   list of person metadata (submitter or originator)
        Returns:
            str    First Name, Last Name; First Name, Last Name; ...
        """
        if not perslist:
            return "MISSING INFORMATION"
        return list_joiner([list_joiner([pers.PS_LAST_NAME, pers.PS_FIRST_NAME],
                                        ',', insert_space=True)
                            for pers in perslist],
                           ';', insert_space=True)

    def gen_org_str(self):
        """
        Generates output string for ORG line.
        Parameter:
            None
        Returns:
            ORG line (str)
        """
        def _comma_escape(elem):
            """
            Helper: escape comma in strings of a list (wrap in double quotes)
            Parameters:
                elem    list of strings
            Returns:
                None
            """
            for i in range(len(elem)):
                if elem[i] and re.search(",", elem[i]):
                    elem[i] = '"' + re.sub('"', '""', elem[i]) +'"'

        def _none_to_empty(elem):
            """
            Helper: substitute None by empty string in a list.
            Parameters:
                elem    list of strings (or None)
            Returns:
                None
            """
            for i in range(len(elem)):
                if elem[i] is None:
                    elem[i] = ''

        elem = [self.metadata.org['OR_CODE'],
                self.metadata.org['OR_NAME'],
                self.metadata.org['OR_ACRONYM'],
                self.metadata.org['OR_UNIT'],
                self.metadata.org['OR_ADDR_LINE1'],
                self.metadata.org['OR_ADDR_LINE2'],
                self.metadata.org['OR_ADDR_ZIP'],
                self.metadata.org['OR_ADDR_CITY'],
                self.metadata.org['OR_ADDR_COUNTRY']] \
            if 'org' in self.metadata and self.metadata.org is not None \
            else [None] * 9
        _none_to_empty(elem)
        _comma_escape(elem)
        return u"{0}, {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}".format(*elem)

    def gen_mname_str(self):
        """
        Generates output string for MNAME line.
        Parameter:
            None
        Returns:
            MNAME line (str)
        """
        return ' '.join(self.metadata.projects)

    def gen_vname_str(self):
        """
        Generates output strings for VNAME lines.
        Parameters:
            None
        Returns:
            generator (str):
                VNAME lines
        """
        yield "end_time of measurement, days from the file reference point"
        for i in range(len(self.variables)):
            comp, flag = self.vname(i)
            yield comp
            if self.variables[i].flagcol:
                if i == 0 or self.variables[i-1].flagcol:
                    yield flag
                else:
                    yield "numflag, no unit"

    def vname(self, vnum):
        """
        Generates a vname string for the measuremnt column and one for a
        possible flag column for variable vnum.
        Parameters:
            vnum   variable index
        Returns:
            tuple (vname_str, vname_str_flag)
        """
        comp = ''
        flag = 'numflag '
        if 'comp_name' in self.variables[vnum].metadata:
            comp = '{}, '.format(
                self.variables[vnum].metadata.comp_name \
                    if self.variables[vnum].metadata.comp_name is not None \
                    else '')
            flag += '{}, no unit'.format(
                self.variables[vnum].metadata.comp_name \
                    if self.variables[vnum].metadata.comp_name is not None \
                    else '')
        else:
            comp = '{}, '.format(
                self.metadata.comp_name if self.metadata.comp_name is not None \
                else '')
            flag += '{}, no unit'.format(
                self.metadata.comp_name if self.metadata.comp_name is not None \
                else '')
        if 'unit' in self.variables[vnum].metadata:
            comp += self.variables[vnum].metadata.unit \
                if self.variables[vnum].metadata.unit != None else ''
        else:
            comp += self.metadata.unit if self.metadata.unit is not None else ''
        # characteristics
        if 'characteristics' in self.variables[vnum].metadata and \
                self.variables[vnum].metadata.characteristics is not None:
            for char in self.variables[vnum].metadata.characteristics.sorted():
                comp += ', {}={}'.format(char.CT_TYPE, char.value_string())
                flag += ', {}={}'.format(char.CT_TYPE, char.value_string())
        for meta in self.internal.ebasmetadata.metadata_list_tag_valuestr(\
                self, vnum):
            value = meta[1]
            if '=' in value:
                value = u'"' + re.sub('"', '""', value) + u'"'
            tag_value = u"{0}={1}".format(meta[0], value)
            if ',' in tag_value:
                tag_value = u'"' + re.sub('"', '""', tag_value) + u'"'
            comp += u", {0}".format(tag_value)
            flag += u", {0}".format(tag_value)
        return (comp, flag)

    def gen_nncom_str(self):
        """
        Generates output strings for NNCOM lines.
        Parameter:
            None
        Returns:
            generator (str):
                NNCOM lines
        """
        # set the lable width:
        lable_width = 29
        for meta in self.internal.ebasmetadata.metadata_list_tag_valuestr(
                self):
            if ':' in meta[1] or '"' in meta[1]:
                val = u'"' + re.sub('"', '""', meta[1]) + u'"'
            else:
                val = meta[1]
            yield u"{0:{width}s} {1}".format(meta[0]+':', val,
                                             width=lable_width)

        yield self.gen_title_str()


    def gen_title_str(self):
        """
        Generates output string for TITLE line.
        Parameter:
            None
        Returns:
            TITLE line (str)
        """
        title = ["starttime", "endtime"]
        for i in range(len(self.variables)):
            title.append(
                self.variables[i].metadata.title \
                    if self.variables[i].metadata.title else '')
            if self.variables[i].flagcol:
                if ('flagtitle' in self.variables[i] and
                    self.variables[i].flagtitle):
                    # flagtitle is an attribute of the variable, not in the
                    # metadata of the variable. This seems more natural, because
                    # it should for example not be stored when reading and
                    # writing files (the flag title should change then because
                    # of possible reorganisations and variable rearangements).
                    # In contrast, the metadata.title should be stored when
                    # reading a file, and then be reused when writing.
                    title.append(self.variables[i].flagtitle)
                elif i == 0 or self.variables[i-1].flagcol:
                    title.append("flag_" + \
                        self.variables[i].metadata.title \
                        if self.variables[i].metadata.title else '')
                else:
                    title.append("flag")
        return " ".join(title)

    def set_data(self, flags=FLAGS_ONE_OR_ALL):
        """
        Sets the NasaAmes1001 data variables.
        Parameter:
            None
        Returns:
            None
        """
        self.nasa1001.DATA = []
        offset = self.metadata.reference_date + self.internal.timeoffset
        # int(x/0.0086400)/10000000.0  --> round to 6 digits (< 1s)
        self.nasa1001.DATA.append(
            [(start - offset).total_seconds() / 86400.
             for start, _ in self.sample_times])
        self.nasa1001.DATA.append(
            [(end - offset).total_seconds() / 86400.0
             for _, end in self.sample_times])
        # restrict x and first dep.var output format to 999.999999 at most
        self.nasa1001.minxmindig = -6
        self.nasa1001.minvmindig = [-6]  # same for endtime
        self.nasa1001.maxvmindig = [None]  # no maximum vmindig for endtime
        self.nasa1001.minvmiss = [True] # use minimal VMISS value for endtime
        self.nasa1001.vmiss = [None]  # no explicit vmiss for end_time
        for j in range(len(self.variables)):
            self.nasa1001.minvmiss.append(False)
            self.nasa1001.maxvmindig.append(None)
            if 'vmiss' in self.variables[j] and self.variables[j].vmiss:
                self.nasa1001.vmiss.append(self.variables[j].vmiss)
            else:
                self.nasa1001.vmiss.append(None)
            if flags == FLAGS_NONE:
                # clean values (set invalid to missing, set 781 to 1/2 value)
                # is done only for export - so the NasaAmes object remains
                # fully valid!
                self.nasa1001.DATA.append(self.values_without_flags(j))
            else:
                self.nasa1001.DATA.append(self.variables[j].values_)
            if self.variables[j].flagcol:
                genflags = self._gen_flags(j)
                flagstrings = [''.join(["{0:03d}".format(flag)
                                        for flag in flags])
                               for flags in genflags]
                flagvar = [float("0." + flagstr)
                           for flagstr in flagstrings]
                # forrce to use n digits after, but at least 3 (if no flags at
                # all)
                flaglen = max([len(flagstr) for flagstr in flagstrings] + [3])
                self.nasa1001.maxvmindig.append(-1 * flaglen)
                # use minimal VMISS value for flags:
                self.nasa1001.minvmiss.append(True)
                if 'flag_vmiss' in self.variables[j] and \
                   self.variables[j].flag_vmiss:
                    if not re.match(r'9\.(999)+', self.variables[j].flag_vmiss):
                        raise ValueError("illegal vmiss for flag: {}".format(
                            self.variables[j].flag_vmiss))
                    self.nasa1001.vmiss.append(self.variables[j].flag_vmiss)
                else:
                    self.nasa1001.vmiss.append(None)
                self.nasa1001.DATA.append(flagvar)

    def _gen_flags(self, vnum):
        """
        Generate the flag sequence to be printed. Take care of this special
        case:
        A flag column applies to multiple data columns, the flag sequence
        contains flag 999 for a sample where one of the other variables
        conatin a valid value (and thus no flag 999).
        This special case has been taken into account for consolidate_flags, but
        the flag sequence itself was NOT changed there (the I/O object should
        remain valid all the time, it should be just done on write).
        """
        # find other variable which use the same flag column:
        othervars = []
        for i in reversed(range(vnum)):
            if self.variables[i].flagcol:
                break
            othervars.append(i)
        if othervars == []:
            # no other variables use the same flag sequence
            return self.variables[vnum].flags
        return [[self.variables[vnum].flags[j][k]
                 for k in range(len(self.variables[vnum].flags[j]))
                 if self.variables[vnum].flags[j][k] != 999 or
                 [self.variables[o].values_[j] for o in othervars
                  if self.variables[o].values_[j] is not None] == []]
                for j in range(len(self.variables[vnum].flags))]



"""
$Id: cc.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for charring correction masterdata

This module implements the class EbasMasterCC.

History:
V.1.0.0  2014-02-25  pe  initial version

"""

from .offline_masterdata import CCOfflineMasterData
from .base import EbasMasterBase

class EbasMasterCC(EbasMasterBase, CCOfflineMasterData):
    """
    Domain Class for charring correction masterdata
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for CC (charring correction)
    # Those are fallback values, will be read from database as soon as possible.
    CCOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read co masterdata if dbh is provided.
        """
        CCOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)




"""
ebas/db_fileindex
$Id: __init__.py 2721 2021-10-22 23:02:49Z pe $

Database I/O for EBAS fileindex database (sqlite3)

This module provides the class IndexDb

History:
V.1.0.0  2015-04-02  pe  initial version

"""

import re

# expose main class to users
from sqlite3 import OperationalError
from .indexdb import IndexDb
from .indexdb_internal import IndexDbInternal



"""
EBAS submission templates and template checks
"""

from logging import getLogger

class NoTemplate(Exception):
    """
    Exception raised when:
     - a specific template does not match (raised in a template class)
     - or no template can be found for a file (raised in get_template)
    """
    pass


class EbasTemplateBase(object):
    """
    Base class for template objects.
    """

    TEMPLATE_NAME = None  # needs to be set in each template

    def __init__(self, fil, ignore_templatecheck=False):
        """
        Initialize object for using it for a specific ebas-io file

        Parameters:
            fil    ebas io file object
        Returns:
            None
        Raises:
            NoTemplate if the template does not match for the file.
        """
        self.file = fil
        self.ignore_templatecheck = ignore_templatecheck
        self.logger = getLogger(self.__class__.__name__)
        self.match()

    def match(self):
        """
        See if this template matches a file.
        Must be implemented in each client class
        """
        raise NotImplementedError()

    def checkfile(self):
        """
        Check the file for template complience.
        """
        # TODO: default checks?
        self.special_checks()

    def special_checks(self):
        """
        Special checks which cannot be done in the defailt checks.
        Must be implemented in each client class
        """
        raise NotImplementedError()

    def error(self, *args, **kwargs):
        """
        Wrapper for self.file.error, adds template name
        """
        if self.ignore_templatecheck:
            self.warning(*args, **kwargs)
        else:
            self.file.error(*args,
                            prefix="Template '{}': ".format(self.TEMPLATE_NAME),
                            **kwargs)

    def warning(self, *args, **kwargs):
        """
        Wrapper for self.file.warning, adds template name
        """
        self.file.warning(*args,
                          prefix="Template '{}': ".format(self.TEMPLATE_NAME),
                          **kwargs)
